import numpy as np
import pytest

import lentil
import lentil.util
import lentil.helper


def test_mesh_nonsquare():
    xx, yy = lentil.helper.mesh((256, 512))
    assert xx.shape == (256, 512)


def test_boundary_slice():
    a = np.zeros((10, 10))
    r, c = np.floor(np.random.uniform(low=0, high=10, size=2)).astype(int)
    a[r:r+3, c:c+3] = 1

    rmin = np.max((r, 0))
    rmax = np.min((r+3, a.shape[0]))
    cmin = np.max((c, 0))
    cmax = np.min((c+3, a.shape[1]))

    slc = lentil.helper.boundary_slice(a)

    assert (slice(rmin, rmax), slice(cmin, cmax)) == slc

@pytest.mark.parametrize('shape, radius', [
    ((256, 256), 256//4),
    ((256, 256), 256//5),
    ((255, 255), 256//4),
    ((255, 255), 256//5)
])
def test_slice_offset(shape, radius):
    shift = np.random.uniform(low=-50, high=50, size=2).astype(int)

    a = lentil.circle(shape=shape, radius=radius, shift=shift, antialias=False)
    slc = lentil.helper.boundary_slice(a)
    offset = lentil.helper.slice_offset(slc, shape=a.shape)

    assert np.all(offset == shift)



import lentil
import numpy as np


def test_propagate_airy(airy2, pupil):
    diameter = 1
    focal_length = 10
    wavelength = 500e-9
    pixelscale = (5e-6, 5e-6)
    shape = (511, 511)
    oversample = 1

    psf_airy = airy2(diameter, focal_length, wavelength, pixelscale, shape, oversample)
    psf_airy = psf_airy/np.max(psf_airy)

    p = pupil(focal_length=focal_length,
              diameter=diameter,
              shape=(512, 512),
              radius=256,
              coeffs=None)

    w = lentil.Wavefront(wavelength=500e-9)
    w *= p
    w = lentil.propagate_fft(w, shape=(511,511), pixelscale=5e-6, oversample=1)
    psf = w.intensity
    psf = psf/np.max(psf)

    np.all(np.isclose(psf, psf_airy, atol=1e-3))


import copy

import numpy as np

import lentil
from lentil import Tilt
import lentil.field
from lentil.field import Field
import lentil.fourier
import lentil.helper

class Wavefront:
    """A class representing a monochromatic wavefront.

    Parameters
    ----------
    wavelength : float
        Wavelength in meters
    pixelscale : float, optional
        Physical sampling of wavefront
    diameter: float, optional
        Wavefront diameter. Default is None
    focal_length : float or np.inf, optional
        Wavefront focal length. A plane wave (default) has an infinite focal
        length (``np.inf``).
    tilt: (2,) array_like, optional
        Radians of wavefront tilt about the x and y axes provided as 
        ``[rx, ry]``. Default is ``[0, 0]`` (no tilt).
    ptype : lentil.ptype, optional
        Plane type. Default is ``lentil.none``.
    
    """
    def __init__(self, wavelength, pixelscale=None, diameter=None, focal_length=None,
                 tilt=None, ptype=None):
        
        #: float: Wavefront focal length
        self.focal_length = focal_length if focal_length else np.inf

        #: float: Wavefront diameter
        self.diameter = diameter

        #: tuple of ints: Wavefront shape
        self.shape = ()

        self._wavelength = wavelength
        self._pixelscale = None if pixelscale is None else np.broadcast_to(pixelscale, (2,))
        self.ptype = lentil.ptype(ptype)

        if tilt is not None:
            if len(tilt) != 2:
                raise ValueError('tilt must be specified as [rx, ry]')
            tilt = [Tilt(x=tilt[0], y=tilt[1])]

        self.data = [Field(data=np.array(1, dtype=complex),
                           offset=None,
                           tilt=tilt)]

    def __mul__(self, plane):
        return plane.multiply(self)

    def __rmul__(self, other):
        return self.__mul__(other)

    @property
    def wavelength(self):
        """Wavefront wavelength
        
        Returns
        -------
        float
        """
        return self._wavelength

    @property
    def pixelscale(self):
        """Physical sampling of wavefront
        
        Returns
        -------
        tuple of floats
        """
        return self._pixelscale

    @property
    def ptype(self):
        """Wavefront plane type
        
        Returns
        -------
        ptype object
        """
        return self._ptype
    
    @ptype.setter
    def ptype(self, value):
        if lentil.ptype(value) not in (lentil.none, lentil.pupil, lentil.image):
            raise TypeError(f"invalid ptype '{value}' for Wavefront")
        else:
            self._ptype = lentil.ptype(value)

    @property
    def field(self):
        """Wavefront complex field
        
        Returns
        -------
        ndarray
        """
        out = np.zeros(self.shape, dtype=complex)
        for field in self.data:
            out = lentil.field.insert(field, out)
        return out

    @property
    def intensity(self):
        """Wavefront intensity
        
        Returns
        -------
        ndarray
        """
        out = np.zeros(self.shape, dtype=float)
        for field in lentil.field.reduce(self.data):
            out = lentil.field.insert(field, out, intensity=True)
        return out

    @classmethod
    def empty(cls, wavelength, pixelscale=None, diameter=None, focal_length=None,
              tilt=None, shape=None, ptype=None):
        """Create an empty Wavefront

        The resulting wavefront will have an empty :attr:`data` attribute.

        Parameters
        ----------
        
        """
        w = cls(wavelength=wavelength, pixelscale=pixelscale, diameter=diameter,
                focal_length=focal_length, tilt=tilt, ptype=ptype)
        w.data = []
        w.shape = () if shape is None else shape
        return w
    

    def insert(self, out, weight=1):
        """Directly insert wavefront intensity data into an output array.

        This method can avoid repeatedly allocating large arrays of zeros
        when accumulating :attr:`intensity`.

        Parameters
        ----------
        out : ndarray
            Array to insert wavefront data into
        weight : float
            Scale factor applied to wavefront data

        Returns
        -------
        out : ndarray
            Array with wavefront data inserted into it at the appropriate location
        """
        for field in lentil.field.reduce(self.data):
            out = lentil.field.insert(field, out, intensity=True, weight=weight)
        return out


def _overlap(field_shape, field_shift, output_shape):
    # Return True if there's any overlap between a shifted field and the
    # output shape
    output_shape = np.asarray(output_shape)
    field_shape = np.asarray(field_shape)
    field_shift = np.asarray(field_shift)

    # Output coordinates of the upper left corner of the shifted data array
    field_shifted_ul = (output_shape / 2) - (field_shape / 2) + field_shift

    if field_shifted_ul[0] > output_shape[0]:
        return False
    if field_shifted_ul[0] + field_shape[0] < 0:
        return False
    if field_shifted_ul[1] > output_shape[1]:
        return False
    if field_shifted_ul[1] + field_shape[1] < 0:
        return False
    return True



import numpy as np
import lentil


def test_power_spectrum_wfe():
    mask = lentil.circle((256, 256), 128, antialias=False)
    rms = 50e-9
    opd = lentil.wfe.power_spectrum(mask, pixelscale=1/256, rms=rms,
                                      half_power_freq=5, exp=3)

    assert np.std(opd[np.nonzero(opd)])/rms >= 0.8


def test_translation_defocus():
    mask = lentil.circle((256, 256), 128, antialias=False)
    f_number = np.random.normal(10)
    translation = np.random.uniform(low=-0.5e-3, high=0.5e-3)

    opd = lentil.wfe.translation_defocus(mask, f_number, translation)
    pv_defocus = translation/(8*f_number**2)

    assert np.isclose(np.abs(pv_defocus), np.abs(np.max(opd) - np.min(opd)))



import numpy as np
from scipy.ndimage import map_coordinates
import lentil


def centroid(img):
    """Compute image centroid location.

    Parameters
    ----------
    img : array_like
        Input array.

    Returns
    -------
    tuple
        ``(r,c)`` centroid location.
    """

    img = np.asarray(img)
    img = img/np.sum(img)
    nr, nc = img.shape
    rr, cc = np.mgrid[0:nr, 0:nc]

    r = np.dot(rr.ravel(), img.ravel())
    c = np.dot(cc.ravel(), img.ravel())

    return r, c


def pad(array, shape):
    """Zero-pad an array.

    Note that pad works for both two and three dimensional arrays.

    Parameters
    ----------
    array : array_like
        Array to be padded.

    shape : tuple of ints
        Shape of output array in ``(nrows, ncols)``.

    Returns
    -------
    padded : ndarray
        Zero-padded array with shape ``(nrows, ncols)``. If ``array`` has a
        third dimension, the return shape will be ``(nrows, ncols, depth)``.
    """

    array = np.asarray(array)

    offset = 0
    if array.ndim == 3:
        offset = 1


    dr = shape[0] - array.shape[0+offset]
    dc = shape[1] - array.shape[1+offset]

    if dr <= 0:
        rmin0 = (array.shape[0+offset] - shape[0])//2
        rmax0 = rmin0 + shape[0]
        rmin1 = 0
        rmax1 = shape[0]
    else:
        rmin0 = 0
        rmax0 = array.shape[0+offset]
        rmin1 = (shape[0] - array.shape[0+offset])//2
        rmax1 = rmin1 + array.shape[0+offset]

    if dc <= 0:
        cmin0 = (array.shape[1+offset] - shape[1])//2
        cmax0 = cmin0 + shape[1]
        cmin1 = 0
        cmax1 = shape[1]
    else:
        cmin0 = 0
        cmax0 = array.shape[1]
        cmin1 = (shape[1] - array.shape[1+offset])//2
        cmax1 = cmin1 + array.shape[1+offset]

    if array.ndim < 3:
        padded = np.zeros((shape[0], shape[1]), dtype=array.dtype)
        padded[rmin1:rmax1, cmin1:cmax1] = array[rmin0:rmax0, cmin0:cmax0]
    else:
        padded = np.zeros((array.shape[0], shape[0], shape[1]), dtype=array.dtype)
        padded[:, rmin1:rmax1, cmin1:cmax1] = array[:, rmin0:rmax0, cmin0:cmax0]

    return padded


def subarray(a, shape, shift=(0,0)):
    """Extract a contiguous subarray from a larger array.

    The subarray is extracted about the center of the source array unless
    a shift is specified.

    Parameters
    ----------
    a : array_like
        Source array
    shape : array_like of ints
        Shape of subarray array in ``(nrows, ncols)``.
    shift : array_like of ints
        Relative shift of the center of the subarray in ``(row, col)``.
    
    Returns
    -------
    out : ndarray
        Subarray extracted from the source array.
    """

    a = np.asarray(a)
    shape = np.asarray(shape)

    rmin = a.shape[0]//2 - shape[0]//2 + shift[0]
    cmin = a.shape[1]//2 - shape[1]//2 + shift[1]
    rmax = rmin + shape[0]
    cmax = cmin + shape[1]

    if any((rmin<0, cmin<0, rmax>a.shape[0], cmax>a.shape[1])):
        raise ValueError('window lies outside of array')

    return a[rmin:rmax, cmin:cmax]


def window(img, shape=None, slice=None):
    """Extract an appropriately sized, potentially windowed array

    Parameters
    ----------
    img : array_like
        Data to window

    shape : array_like or None, optional
        Output shape given as (nrows, ncols). If ``None`` (default), an
        uncropped array is returned.

    slice : array_like or None, optional
        Indices of ``img`` array to return given as (r_start, r_end, c_start,
        c_end). This definition follows standard numpy indexing.

    Returns
    -------
    data : ndarray
        Trimmed and windowed ``img`` array

    Notes
    -----
    * If ``img`` is a single value (img.size == 1), self.data is returned
      regardless of what ``shape`` and ``slice`` are.
    * If ``shape`` is given but ``slice`` is ``None``, the returned ndarray
      is trimmed about the center of the array using :func:`lentil.pad`.
    * If ``slice`` is given but ``shape`` is ``None``, the returned ndarray
      is extracted from ``img`` according to the indices in ``slice``
    * If both ``shape`` and ``slice`` are given, the returned ndarray is
      extracted from ``img`` according to the indices in ``slice`` and the
      following expressions must also be true:

    .. code::

        shape[0] = (slice[1] - slice[0]) = (r_end - r_start)
        shape[1] = (slice[3] - slice[2]) = (c_end - c_start)

    """

    img = np.asarray(img)
    if img.size == 1:
        return img

    if shape is None and slice is None:
        return img
    elif slice is not None:
        if shape is not None:
            # ensure size consistency
            assert(slice[1] - slice[0]) == shape[0]
            assert(slice[3] - slice[2]) == shape[1]

        # return the requested view. Note that numpy will implicitly handle a
        # third dimension if one is present
        return img[slice[0]:slice[1], slice[2]:slice[3]]

    else:
        # return the padded array. Note that pad will handle a third dimension
        # if one exists
        return lentil.pad(img, shape)


def boundary(x, threshold=0):
    """Find bounding row and column indices of data within an array.

    Parameters
    ----------
    x : array_like
        Input array

    threshold : float, optional
        Masking threshold to apply before boundary finding. Only values
        in x that are larger than threshold are considered in the boundary
        finding operation. Default is 0.

    Returns
    -------
    rmin, rmax, cmin, cmax : ints
        Boundary indices

    """
    x = np.asarray(x)
    x = (x > threshold)

    rows = np.any(x, axis=1)
    cols = np.any(x, axis=0)

    rmin, rmax = np.where(rows)[0][[0, -1]]
    cmin, cmax = np.where(cols)[0][[0, -1]]

    return rmin, rmax, cmin, cmax


def rebin(img, factor):
    """Rebin an image by an integer factor.

    Parameters
    ----------
    img : array_like
        Array or cube of arrays to rebin. If a cube is provided, the first dimension
        should index the image slices.

    factor : int
        Rebinning factor

    Returns
    -------
    img : ndarray
        Rebinned image

    See Also
    --------
    :func:`rescale`

    """
    img = np.asarray(img)

    if np.iscomplexobj(img):
        raise ValueError('rebin is not defined for complex data')

    if img.ndim == 3:
        rebinned_shape = (img.shape[0], img.shape[1]//factor, img.shape[2]//factor)
        img_rebinned = np.zeros(rebinned_shape, dtype=img.dtype)
        for i in range(img.shape[0]):
            img_rebinned[i] = img[i].reshape(rebinned_shape[1], factor,
                                             rebinned_shape[2], factor).sum(-1).sum(1)
    else:
        img_rebinned = img.reshape(img.shape[0]//factor, factor, img.shape[1]//factor,
                                   factor).sum(-1).sum(1)

    return img_rebinned


def rescale(img, scale, shape=None, mask=None, order=3, mode='nearest',
            unitary=True):
    """Rescale an image by interpolation.

    Parameters
    ----------
    img : array_like
        Image to rescale

    scale : float
        Scaling factor. Scale factors less than 1 will shrink the image. Scale
        factors greater than 1 will grow the image.

    shape : array_like or int, optional
        Output shape. If None (default), the output shape will be the input img
        shape multiplied by the scale factor.

    mask : array_like, optional
        Binary mask applied after rescaling. If None (default), a mask is
        created from the nonzero portions of img. To skip masking operation,
        set ``mask = np.ones_like(img)``

    order : int, optional
        Order of spline interpolation used for rescaling operation. Default is
        3. Order must be in the range 0-5.

    mode : {'constant', 'nearest', 'reflect', 'wrap'}, optional
        Points outside the boundaries of the input are filled according to the
        given mode. Default is 'constant'.

    unitary : bool, optional
        Normalization flag. If True (default), a normalization is performed on
        the output such that the rescaling operation is unitary and image power
        (if complex) or intensity (if real) is conserved.

    Returns
    -------
    img : ndarray
        Rescaled image.

    Notes
    -----
    The post-rescale masking operation should have no real effect on the
    resulting image but is included to eliminate interpolation artifacts that
    sometimes appear in large clusters of zeros in rescaled images.

    See Also
    --------
    :func:`rebin`

    """

    img = np.asarray(img)

    if mask is None:
        # take the real portion to ensure that even if img is complex, mask will
        # be real
        mask = np.zeros_like(img).real
        mask[img != 0] = 1

    if shape is None:
        shape = np.ceil((img.shape[0]*scale, img.shape[1]*scale)).astype(int)
    else:
        if np.isscalar(shape):
            shape = np.ceil((shape*scale, shape*scale)).astype(int)
        else:
            shape = np.ceil((shape[0]*scale, shape[1]*scale)).astype(int)

    x = (np.arange(shape[1], dtype=np.float64) - shape[1]/2.)/scale + img.shape[1]/2.
    y = (np.arange(shape[0], dtype=np.float64) - shape[0]/2.)/scale + img.shape[0]/2.

    xx, yy = np.meshgrid(x, y)

    mask = map_coordinates(mask, [yy, xx], order=1, mode='nearest')
    mask[mask < np.finfo(mask.dtype).eps] = 0

    if np.iscomplexobj(img):
        out = np.zeros(shape, dtype=np.complex128)
        out.real = map_coordinates(img.real, [yy, xx], order=order, mode=mode)
        out.imag = map_coordinates(img.imag, [yy, xx], order=order, mode=mode)
    else:
        out = map_coordinates(img, [yy, xx], order=order, mode=mode)

    if unitary:
        out *= np.sum(img)/np.sum(out)

    out *= mask

    return out


def pixelscale_nyquist(wave, f_number):
    """Compute the output plane sampling which is Nyquist sampled for
    intensity.

    Parameters
    ----------
    wave : float
        Wavelength in meters

    f_number : float
        Optical system F/#

    Returns
    -------
    float
        Sampling in meters

    """
    return f_number * wave / 2


def min_sampling(wave, z, du, shape, min_q):
    num = np.min(wave) * z
    return num/(min_q * du[0] * shape[0]), num/(min_q * du[1] * shape[1])


def normalize_power(array, power=1):
    r"""Normalizie the power in an array.

    The total power in an array is given by

    .. math::

        P = \sum{\left|\mbox{array}\right|^2}

    A normalization coefficient is computed as

    .. math::

        c = \sqrt{\frac{p}{\sum{\left|\mbox{array}\right|^2}}}

    The array returned by a will be scaled by the normalization coefficient so
    that its power is equal to :math:`p`.

    Parameters
    ----------
    array : array_like
        Array to be normalized

    power : float, optional
        Desired power in normalized array. Default is 1.

    Returns
    -------
    array : ndarray
        Normalized array

    """
    array = np.asarray(array)
    return array * np.sqrt(power/np.sum(np.abs(array)**2))


def sanitize_shape(shape):
    shape = np.asarray(shape)
    if shape.size == 0:
        shape = ()
    else:
        if shape.shape == ():
            shape = np.append(shape, shape)
    return tuple(shape)


def sanitize_bandpass(vec):
    vec = np.asarray(vec)
    if vec.shape == ():
        vec = vec[np.newaxis, ...]
    return vec



import numpy as np

import lentil.fourier
import lentil.helper


def test_dft2_even():
    n = 10
    f = np.random.rand(n, n) + 1j * np.random.rand(n, n)

    F_dft = lentil.fourier.dft2(f, 1/n, unitary=False)
    F_fft = np.fft.fftshift(np.fft.fft2(np.fft.ifftshift(f)))

    assert np.allclose(F_dft, F_fft)


def test_idft2_even():
    n = 10
    F = np.random.rand(n, n) + 1j * np.random.rand(n, n)

    f_dft = lentil.fourier.idft2(F, 1/n, unitary=False)
    f_fft = np.fft.fftshift(np.fft.ifft2(np.fft.ifftshift(F)))

    assert np.allclose(f_dft, f_fft)


def test_dft2_odd():
    n = 11
    f = np.random.rand(n, n) + 1j * np.random.rand(n, n)

    F_dft = lentil.fourier.dft2(f, 1/n, unitary=False)
    F_fft = np.fft.fftshift(np.fft.fft2(np.fft.ifftshift(f)))

    assert np.allclose(F_dft, F_fft)


def test_idft2_odd():
    n = 11
    F = np.random.rand(n, n) + 1j * np.random.rand(n, n)

    f_dft = lentil.fourier.idft2(F, 1/n, unitary=False)
    f_fft = np.fft.fftshift(np.fft.ifft2(np.fft.ifftshift(F)))

    assert np.allclose(f_dft, f_fft)


def test_dft2_inverse():
    n = 10
    f = np.random.rand(n, n) + 1j * np.random.rand(n, n)

    F_dft = lentil.fourier.dft2(f, 1/n, unitary=False)
    f_dft = lentil.fourier.idft2(F_dft, 1/n, unitary=False)

    assert np.allclose(f, f_dft)


def test_dft2_unitary():
    n = 10
    f = np.random.rand(n, n) + 1j * np.random.rand(n, n)
    f_power = np.sum(np.abs(f)**2)

    F = lentil.fourier.dft2(f, 1/n, unitary=True)
    F_power = np.sum(np.abs(F)**2)

    assert np.allclose(f_power, F_power)


def test_dft2_shift():
    n = 100
    shift = np.round(np.random.uniform(low=-25, high=25, size=2))
    f = np.ones((n, n))

    F = lentil.fourier.dft2(f, 1/n, shift=shift)

    (xc, yc) = (np.floor(n/2), np.floor(n/2))
    (x, y) = lentil.util.centroid(np.abs(F)**2)
    observed_shift = (x-xc, y-yc)
    assert np.array_equal(shift, observed_shift)


def test_dft2_offset():
    n = 10
    m = 3

    f = np.zeros((n,n), dtype=complex)
    r,c = np.random.randint(0, n-m, size=2)
    f[r:r+m, c:c+m] = np.random.rand(m, m) + 1j * np.random.rand(m,m)
    slc = lentil.helper.boundary_slice(f)
    offset = lentil.helper.slice_offset(slc, f.shape)

    F = lentil.fourier.dft2(f, alpha=1/m, shape=10)
    FF = lentil.fourier.dft2(f[slc], alpha=1/m, shape=10, offset=offset)

    assert np.allclose(F, FF)


def test_dft2_out():
    n = 10
    f = np.random.normal(size=(n,n)).astype(complex)
    F = lentil.fourier.dft2(f, 1/n, out=f)
    assert f is F


def test_dft2_rect():
    m, n = 10, 11
    f = np.random.rand(m, n) + 1j * np.random.rand(m, n)

    F_dft = lentil.fourier.dft2(f, [1/m, 1/n], unitary=False)
    F_fft = np.fft.fftshift(np.fft.fft2(np.fft.ifftshift(f)))

    assert np.allclose(F_dft, F_fft)




import pytest
import numpy as np
import lentil


def test_zernike_rho_theta():
    with pytest.raises(ValueError):
        lentil.zernike(mask=1, index=1, normalize=True, rho=1, theta=None)


def test_zernike_basis():
    basis = lentil.zernike_basis(mask=np.ones((3, 3)), modes=1, vectorize=False)
    assert np.array_equal(basis, np.ones((1, 3, 3)))


def test_zernike_basis_vectorize():
    basis = lentil.zernike_basis(mask=np.ones((3, 3)), modes=1, vectorize=True)
    assert np.array_equal(basis, np.ones((1, 9)))


def test_zernike_fit():
    mask = lentil.circle((256, 256), 128, antialias=False)
    coeffs = np.random.rand(4)*100e-9
    opd = lentil.zernike_compose(mask, coeffs)
    fit_coeffs = lentil.zernike_fit(opd, mask, np.arange(1, 5))
    assert np.all(np.isclose(coeffs, fit_coeffs))


def test_zernike_remove():
    mask = lentil.circle((256, 256), 128, antialias=False)
    coeffs = np.random.rand(4)*100e-9
    opd = lentil.zernike_compose(mask, coeffs)
    residual = lentil.zernike_remove(opd, mask, np.arange(1, 5))
    assert np.all(np.isclose(residual, np.zeros_like(residual)))

def test_zernike_random_mask():
    mask = lentil.circle((256, 256), 128, antialias=False)
    random_mask = np.random.uniform(low=-1, high=1, size=mask.shape) * mask

    assert np.array_equal(lentil.zernike(mask, 1), lentil.zernike(random_mask, 1))



import pytest
import numpy as np
import lentil


def test_default_wavefront():
    w = lentil.Wavefront(wavelength=500e-9)
    assert np.array_equal(w.field, 1+0j)
    assert np.array_equal(w.intensity, 1)

def test_wavefront_rmul():
    w = lentil.Wavefront(wavelength=500e-9)
    p = lentil.Plane()
    assert p * w

def test_wavefront_propagate_image_non_pupil():
    w = lentil.Wavefront(wavelength=500e-9)
    with pytest.raises(TypeError):
        lentil.propagate_dft(w, shape=(64,64), pixelscale=5e-6)


@pytest.mark.parametrize('field_shape, field_shift, output_shape', [
    ((5,5), (-25,0), (10,10)),
    ((5,5), (25,0), (10,10)),
    ((5,5), (0,-25), (10,10)),
    ((5,5), (0,25), (10,10))
])
def test_overlap(field_shape, field_shift, output_shape):
    assert lentil.wavefront._overlap(field_shape, field_shift, output_shape) is False



import numpy as np

import lentil


def circle(shape, radius, shift=(0, 0), antialias=True):
    """Draw a circle

    Parameters
    ----------
    shape : array_like
        Size of output in pixels (nrows, ncols)
    radius : float
        Radius of circle in pixels
    shift : (2,) array_like, optional
        How far to shift center in float (rows, cols). Default is (0, 0).
    antialias : bool, optional
        If True (default), the shape edges are antialiased.

    Returns
    -------
    ndarray

    """
    rr, cc = lentil.helper.mesh(shape)
    r = np.sqrt(np.square(rr - shift[0]) + np.square(cc - shift[1]))
    mask = np.clip(radius + 0.5 - r, 0.0, 1.0)
    if not antialias:
        mask[mask > 0] = 1
    return mask


def hexagon(shape, radius, shift=(0, 0), rotate=False, antialias=True):
    """Draw a hexagon

    Parameters
    ----------
    shape : array_like
        Size of output in pixels (nrows, ncols)
    radius : float
        Radius of outscribing circle (which also equals the side length) in
        pixels.
    shift : tuple of floats, optional
        How far to shift center in (rows, cols). Default is (0, 0).
    rotate : bool, optional
        Rotate mask so that flat sides are aligned with the Y direction instead
        of the default orientation which is aligned with the X direction.
    antialias : bool, optional
        If True (default), the shape edges are antialiased.
        
    Returns
    -------
    ndarray

    """
    shape = np.broadcast_to(shape, (2,))
    r, c = lentil.helper.mesh(shape, shift)
    mask = np.ones(shape)

    inner_radius = radius * np.sqrt(3)/2

    for n in range(6):
    
        theta = n * np.pi/3 if rotate else n * np.pi/3 + np.pi/6
        rho = r * np.sin(theta) + c * np.cos(theta)
    
        if antialias:
            slc = np.clip(inner_radius + 0.5 - rho, 0.0, 1.0)
        else:
            slc = np.ones(shape)
            slc[rho > inner_radius] = 0
    
        mask = np.minimum(mask, slc)

    return mask


def rectangle(shape, width, height, shift=(0,0), angle=0, antialias=True):
    """Draw a rectangle

    Parameters
    ----------
    shape : array_like
        Size of output in pixels (nrows, ncols)
    width : float
        Width of rectangle in pixels
    height : float
        Height of rectangle in pixels
    shift : tuple of floats, optional
        How far to shift center in (rows, cols). Default is (0, 0).
    angle : float, optional
        Rotation of rectangle in degrees counterclockwise from horizontal.
        Default is 0.
    antialias : bool, optional
        If True (default), the shape edges are antialiased.
        
    Returns
    -------
    ndarray
    
    """
    shape = np.broadcast_to(shape, (2,))
    rr, cc = lentil.helper.mesh(shape, shift, angle)
    rect = np.ones(shape)

    width_clip = np.clip(0.5 + (width/2) - np.abs(cc), 0, 1)
    height_clip = np.clip(0.5 + (height/2) - np.abs(rr), 0, 1)

    rect = np.minimum(np.minimum(rect, width_clip), height_clip)

    if not antialias:
        rect[rect > 0] = 1

    return rect


def spider(shape, width, angle=0, shift=(0,0), antialias=True):
    """Draw a spider

    Parameters
    ----------
    shape : array_like
        Size of output in pixels (nrows, ncols)
    width : float
        Width of rectangle in pixels
    angle : float, optional
        Rotation of spider in degrees counterclockwise from horizontal.
        Default is 0.
    shift : tuple of floats, optional
        How far to shift center in (rows, cols). Default is (0, 0).
    antialias : bool, optional
        If True (default), the spider edges are antialiased.
    
    Returns
    -------
    ndarray

    """
    shape = np.broadcast_to(shape, (2,))
    len = np.sqrt(2) * np.max(shape)/2  # max length when angle is a multiple of 45 deg
    shift_dist = len / 2
    shift_row = -shift_dist * np.sin(np.deg2rad(angle))
    shift_col = shift_dist * np.cos(np.deg2rad(angle))
    shift = (shift[0] + shift_row, shift[1] + shift_col)
    return 1 - lentil.rectangle(shape, len, width, shift, angle, antialias)


import numpy as np

import lentil


def test_blackbody_planck_radiance():
    b = lentil.radiometry.Blackbody(wave=np.arange(500, 5000), temp=3000, waveunit='nm',
                                    valueunit='photlam')

    planck_radiance = lentil.radiometry.planck_radiance(wave=np.arange(500, 5000),
                                                        temp=3000, waveunit='nm',
                                                        valueunit='photlam')

    assert np.array_equal(b.value, planck_radiance)


def test_blackbody_add():
    b = lentil.radiometry.Blackbody(wave=np.arange(500, 5000), temp=3000, waveunit='nm',
                                    valueunit='wlam')
    r = b + 10
    assert isinstance(r, lentil.radiometry.Spectrum)
    assert not isinstance(r, lentil.radiometry.Blackbody)
    assert np.array_equal(r.value, b.value + 10)


def test_blackbody_mul():
    b = lentil.radiometry.Blackbody(wave=np.arange(500, 5000), temp=3000, waveunit='nm',
                                    valueunit='photlam')
    r = b * 3
    assert isinstance(r, lentil.radiometry.Spectrum)
    assert not isinstance(r, lentil.radiometry.Blackbody)
    assert np.array_equal(r.value, b.value * 3)



import numpy as np

import lentil
import lentil.extent
from lentil.field import Field
from lentil.wavefront import Wavefront


def propagate_fft(wavefront, pixelscale, shape=None, oversample=2, 
                  scratch=None):
    """Propagate a Wavefront in the far-field using the FFT.
    
    Parameters
    ----------
    wavefront : :class:`~lentil.Wavefront`
        Wavefront to propagate
    pixelscale : float or (2,) float
        Physical sampling of output Wavefront. If a single value is supplied,
        the output is assumed to be uniformly sampled in both x and y.
    shape : int or (2,) tuple of ints or None
        Shape of output Wavefront. If None (default), the wavefront shape is
        used.
    oversample : float, optional
        Number of times to oversample the output plane. Default is 2.
    scratch : complex ndarray, optional
        A pre-allocated array used for padding. Providing a sufficiently
        large scratch array can improve broadband propagation performance by 
        avoiding the repeated allocation of arrays of zeros.

    Returns
    -------
    wavefront : :class:`~lentil.Wavefront`
        The propagated Wavefront

    """
    if _has_tilt(wavefront):
        raise NotImplementedError('propagate_fft does not support Wavefronts '
                                  'with fitted tilt. Use propagate_dft instead.')

    ptype_out = _propagate_ptype(wavefront.ptype, method='fraunhofer')
    pixelscale = np.broadcast_to(pixelscale, (2,))
    fft_shape, prop_wavelength = _fft_shape(wavefront.pixelscale, 
                                            pixelscale, 
                                            wavefront.focal_length, 
                                            wavefront.wavelength, 
                                            oversample)
    
    # TODO: verify field_shape is smaller than fft_shape
    # TODO: what if field is bigger than fft_shape?
    # field_shape = wavefront.shape

    if shape is None:
        shape_out = tuple(fft_shape)
        shape = (fft_shape[0]//oversample, fft_shape[1]//oversample)
    else:
        shape = tuple(np.broadcast_to(shape, (2,)))
        if np.any(shape > fft_shape/oversample):
            raise ValueError(f'requested shape {tuple(shape)} is larger in at '
                            f'least one dimension than maximum propagation '
                            f'shape {tuple(fft_shape//oversample)}')
        else:
            shape_out = (shape[0] * oversample, shape[1]*oversample)

    out = Wavefront.empty(wavelength=prop_wavelength,
                          pixelscale = pixelscale/oversample,
                          focal_length=wavefront.focal_length,
                          shape = shape_out,
                          ptype = ptype_out)
    
    if scratch is not None:
        if not all(np.asarray(scratch.shape) > fft_shape):
             raise ValueError(f'scratch must have shape greater than or '
                              f'equal to {tuple(fft_shape)}')

        # zero out the portion of scratch that we're going to use for the
        # propagation and then insert the Wavefront field(s) into scratch
        scratch[0:fft_shape[0], 0:fft_shape[1]] = 0
        for field in wavefront.data:
            scratch[0:fft_shape[0], 0:fft_shape[1]] = lentil.field.insert(field, scratch[0:fft_shape[0], 0:fft_shape[1]])
        field =_fft2(scratch[0:fft_shape[0], 0:fft_shape[1]])

    else:
        field = lentil.pad(wavefront.field, fft_shape)
        field = _fft2(field)
    
    out.data.append(Field(data=field, pixelscale=pixelscale/oversample))

    return out


def scratch_shape(wavelength, dx, du, z, oversample):
    """Compute the scratch shape required for an FFT propagation
    
    Parameters
    ----------
    wavelength : float or array_like
        Wavelength or list of wavelengths
    dx : float or (2,) float
        Physical sampling of input Plane. If a single value is supplied,
        the input is assumed to be uniformly sampled in both x and y.
    du : float or (2,) float
        Physical sampling of output Wavefront. If a single value is supplied,
        the output is assumed to be uniformly sampled in both x and y.
    z : float
        Propagation distance
    oversample : float
        Number of times to oversample the output plane.

    Returns
    -------
    shape : tuple
        Scratch shape

    """
    dx = np.broadcast_to(dx, (2,))
    du = np.broadcast_to(du, (2,))
    fft_shape, _ = _fft_shape(dx, du, z, np.max(wavelength), oversample)
    return tuple(fft_shape)


def _dft_alpha(dx, du, wavelength, z, oversample):
    return ((dx[0]*du[0])/(wavelength*z*oversample),
            (dx[1]*du[1])/(wavelength*z*oversample))


def _fft_shape(dx, du, z, wavelength, oversample):
    # Compute pad shape to satisfy requested sampling. Propagation wavelength
    # is recomputed to account for integer padding of input plane
    alpha = _dft_alpha(dx, du, z, wavelength, oversample)
    fft_shape = np.round(np.reciprocal(alpha)).astype(int)       
    prop_wavelength = np.min((fft_shape/oversample * dx * du)/z)
    return fft_shape, prop_wavelength


def _fft2(x):
    return np.fft.ifftshift(np.fft.fft2(np.fft.fftshift(x), norm='ortho'))


def _has_tilt(wavefront):
    # Return True if and Wavefront Field has nonempty tilt
    for field in wavefront.data:
        if field.tilt:
            return True
    return False


def propagate_dft(wavefront, pixelscale, shape=None, prop_shape=None, 
                  oversample=2, mask=None):
    """Propagate a Wavefront in the far-field using the DFT.

    Parameters
    ----------
    wavefront : :class:`~lentil.Wavefront`
        Wavefront to propagate
    pixelscale : float or (2,) float
        Physical sampling of output Wavefront. If a single value is supplied,
        the output is assumed to be uniformly sampled in both x and y.
    shape : int or (2,) tuple of ints or None
        Shape of output Wavefront. If None (default), the wavefront shape is
        used.
    prop_shape : int or (2,) tuple of ints, optional
        Shape of propagation output. If None (default),
        ``prop_shape = prop``. If ``prop_shape != prop``, the propagation
        result is placed in the appropriate location in the output plane.
        ``prop_shape`` should not be larger than ``prop``.
    oversample : float, optional
        Number of times to oversample the output plane. Default is 2.

    Returns
    -------
    wavefront : :class:`~lentil.Wavefront`
        The propagated Wavefront
    """
    
    ptype_out = _propagate_ptype(wavefront.ptype, method='fraunhofer')
    
    shape = np.asarray(wavefront.shape) if shape is None else np.broadcast_to(shape, (2,))
    prop_shape = np.asarray(shape) if prop_shape is None else np.broadcast_to(prop_shape, (2,))
    shape_out = shape * oversample
    prop_shape_out = prop_shape * oversample

    if mask is not None:
        mask = np.asarray(mask)
        if np.all(mask.shape != shape_out):
            raise ValueError(f'shape mismatch: mask shape {mask.shape} != output shape {tuple(shape_out)}')
        mask_shape = _mask_shape(mask, threshold=0)
        mask_shift = _mask_shift(mask, threshold=0)
        out_extent = lentil.extent.array_extent(mask_shape, mask_shift)
    else:
        out_extent = lentil.extent.array_extent(shape_out, shift=(0,0))

    dx = wavefront.pixelscale
    du = np.broadcast_to(pixelscale, (2,))
    z = wavefront.focal_length

    data = wavefront.data

    out = Wavefront.empty(wavelength=wavefront.wavelength,
                          pixelscale = du/oversample,
                          focal_length=wavefront.focal_length,
                          shape = shape_out,
                          ptype = ptype_out)
        
    for field in data:
        # compute the field shift from any embedded tilts. note the return value
        # is specified as (r, c)
        shift = field.shift(z=wavefront.focal_length, wavelength=wavefront.wavelength,
                            pixelscale=du, oversample=oversample,
                            indexing='ij')
        
        fix_shift = np.fix(shift)
        subpx_shift = shift - fix_shift

        prop_extent = lentil.extent.array_extent(prop_shape_out, fix_shift)

        if lentil.extent.intersect(out_extent, prop_extent):
            intersect_shape = lentil.extent.intersection_shape(out_extent, prop_extent)
            intersect_shift = lentil.extent.intersection_shift(out_extent, prop_extent)
            intersect_extent = lentil.extent.array_extent(intersect_shape, intersect_shift)

            # compute additional shift rquired to offset any output clipping that
            # may have occurred due to a mask or extending beyond the full output
            # shape. 
            # NOTE: It appears the same functionality is available using 
            # extent.intersection_shift() but there is an off by one error on only
            # one of the shift dimensions for some reason that causes the tests to 
            # fail
            prop_center = lentil.extent.array_center(prop_extent)
            intersect_center = lentil.extent.array_center(intersect_extent)
            prop_shift = np.array(prop_center) - np.array(intersect_center)

            alpha = _dft_alpha(dx=dx, du=du, z=z,
                               wavelength=wavefront.wavelength,
                               oversample=oversample)
            data = lentil.fourier.dft2(f=field.data, alpha=alpha,
                                       shape=intersect_shape,
                                       shift=prop_shift + subpx_shift,
                                       offset=field.offset, unitary=True)
            out.data.append(Field(data=data, pixelscale=du / oversample,
                                  offset=intersect_shift))

    return out


def _mask_shape(x, threshold=0):
    # compute the shape of a masked area inside an array of zeros
    rmin, rmax, cmin, cmax = lentil.boundary(x, threshold)
    return rmax - rmin + 1, cmax - cmin + 1


def _mask_shift(x, threshold=0):
    # compute the shift of a masked area inside an array of zeros
    shape_full = x.shape
    rc_full, cc_full = shape_full[0]//2, shape_full[1]//2

    rmin_extent, rmax_extent, cmin_extent, cmax_extent = lentil.boundary(x, threshold)
    shape_extent = (rmax_extent-rmin_extent+1, cmax_extent-cmin_extent+1)
    rc_extent, cc_extent = rmin_extent + shape_extent[0]//2, cmin_extent + shape_extent[1]//2

    return rc_extent - rc_full, cc_extent - cc_full


def _propagate_ptype(ptype, method='fraunhofer'):
    if method == 'fraunhofer':
        if ptype not in (lentil.pupil, lentil.image):
            raise TypeError("Wavefront must have ptype 'pupil' "\
                        "or 'image'")
        
        if ptype == lentil.pupil:
            return lentil.image
        else:
            return lentil.pupil


import numpy as np
import pytest

import lentil.field
from lentil.field import Field

@pytest.mark.parametrize('a_data, a_offset, b_data, b_offset, data, offset', [
    (1, [0,0], 1, [0,0], 1+0j, [0,0]),
    (1, [0,0], 1, [1,0], [], [0,0]),
    (np.ones((2,2)), [0,0], np.ones((2,2)), [0,0], np.ones((2,2)), [0,0]),
    (np.ones((2,2)), [0,0], np.ones((2,2)), [4,4], [], [0,0]),
    (1, [0,0], np.ones((3,3)), [-5,-5], np.ones((3,3)), [-5,-5]),
    (np.ones((5,4)), [-2,-2], np.ones((3,3)), [0,-1], np.ones((2,2)), [0,-1])

])
def test_multiply(a_data, a_offset, b_data, b_offset, data, offset):
    a = Field(data=a_data, pixelscale=1, offset=a_offset)
    b = Field(data=b_data, pixelscale=1, offset=b_offset)
    c = a * b
    assert np.array_equal(c.data, data)
    assert np.array_equal(c.offset, offset)

def test_multiply_disjoint():
    a = Field(data=1, pixelscale=1, offset=[0,0])
    b = Field(data=1, pixelscale=1, offset=[-10,0])
    c = a * b
    assert c.size == 0


def test_merge():
    a = Field(data=np.ones((5,4)), pixelscale=1, offset=[-2,-2])
    b = Field(data=np.ones((3,3)), pixelscale=1, offset=[0,-1])
    c = lentil.field.merge(a, b)

    c_data = np.zeros((6,5))
    c_data[0:5,0:4] += 1
    c_data[3:6,2:5] += 1

    assert np.array_equal(c.data, c_data)
    assert np.array_equal(c.offset, [-1, -2])



import pytest
import numpy as np

from lentil.radiometry import Spectrum


def test_spectrum_wave_value_shape_match():
    wave = [1, 2, 3]
    value = [4, 5, 6, 7]
    with pytest.raises(ValueError):
        s = Spectrum(wave, value)


def test_spectrum_add_scalar():
    s = Spectrum([1, 2, 3], [1, 1, 1])
    r = s + 5
    assert np.array_equal(r.wave, s.wave)
    assert np.array_equal(r.value, [6, 6, 6])


def test_spectrum_add_spectrum():
    s1 = Spectrum([1, 2, 3], [1, 1, 1])
    s2 = Spectrum([1, 2, 3], [2, 3, 4])
    r = s1 + s2
    assert np.array_equal(r.wave, s1.wave)
    assert np.array_equal(r.value, [3, 4, 5])


def test_spectrum_add_array():
    s = Spectrum([1, 2, 3], [1, 1, 1])
    r = s + [5, 6, 7]
    assert np.array_equal(r.wave, s.wave)
    assert np.array_equal(r.value, [6, 7, 8])


def test_spectrum_subtract_scalar():
    s = Spectrum([1, 2, 3], [3, 3, 3])
    r = s - 2
    assert np.array_equal(r.wave, s.wave)
    assert np.array_equal(r.value, [1, 1, 1])


def test_spectrum_subtract_spectrum():
    s1 = Spectrum([1, 2, 3], [2, 3, 4])
    s2 = Spectrum([1, 2, 3], [1, 1, 1])
    r = s1 - s2
    assert np.array_equal(r.wave, s1.wave)
    assert np.array_equal(r.value, [1, 2, 3])


def test_spectrum_subtract_array():
    s = Spectrum([1, 2, 3], [5, 6, 7])
    r = s - [1, 1, 1]
    assert np.array_equal(r.wave, s.wave)
    assert np.array_equal(r.value, [4, 5, 6])


def test_spectrum_mul_scalar():
    s = Spectrum([1, 2, 3], [1, 2, 3])
    r = s * 5
    assert np.array_equal(r.wave, s.wave)
    assert np.array_equal(r.value, [5, 10, 15])


def test_spectrum_mul_spectrum():
    s1 = Spectrum([1, 2, 3], [2, 2, 2])
    s2 = Spectrum([1, 2, 3], [2, 3, 4])
    r = s1 * s2
    assert np.array_equal(r.wave, s1.wave)
    assert np.array_equal(r.value, [4, 6, 8])


def test_spectrum_mul_array():
    s = Spectrum([1, 2, 3], [2, 2, 2])
    r = s * [5, 6, 7]
    assert np.array_equal(r.wave, s.wave)
    assert np.array_equal(r.value, [10, 12, 14])


def test_spectrum_truediv_scalar():
    s = Spectrum([5, 10, 15], [5, 10, 15])
    r = s / 5
    assert np.array_equal(r.wave, s.wave)
    assert np.array_equal(r.value, [1, 2, 3])


def test_spectrum_truediv_spectrum():
    s1 = Spectrum([1, 2, 3], [4, 6, 8])
    s2 = Spectrum([1, 2, 3], [2, 3, 4])
    r = s1 / s2
    assert np.array_equal(r.wave, s1.wave)
    assert np.array_equal(r.value, [2, 2, 2])


def test_spectrum_truediv_array():
    s = Spectrum([1, 2, 3], [10, 12, 14])
    r = s / [5, 6, 7]
    assert np.array_equal(r.wave, s.wave)
    assert np.array_equal(r.value, [2, 2, 2])


def test_spectrum_rmul_scalar():
    s = Spectrum([1, 2, 3], [1, 2, 3])
    r = 5 * s
    assert np.array_equal(r.wave, s.wave)
    assert np.array_equal(r.value, [5, 10, 15])


def test_spectrum_pow():
    s = Spectrum([1, 2, 3], [2, 3, 4])
    r = s**2
    assert(np.array_equal(r.wave, s.wave))
    assert(np.array_equal(r.value, s.value**2))


def test_spectrum_sample():
    s = Spectrum([1, 2, 3], [2, 3, 4])
    r = s.sample([1.5, 2.5])
    assert np.array_equal(r, [2.5, 3.5])


def test_spectrum_resample():
    s = Spectrum([1, 2, 3], [2, 3, 4])
    s.resample([1.5, 2.5])
    assert np.array_equal(s.wave, [1.5, 2.5])
    assert np.array_equal(s.value, [2.5, 3.5])


def test_spectrum_asarray():
    s = Spectrum([1, 2, 3], [2, 3, 4])
    r = s.asarray()
    assert np.array_equal(r, np.array([[1, 2, 3], [2, 3, 4]]))


def test_spectrum_bin_preserve_power_true():
    s = Spectrum([1, 2, 3, 4, 5], [2, 2, 2, 2, 2])
    assert np.array_equal(s.bin([2, 3]), [0.5, 1.5])


def test_spectrum_bin_preserve_power_false():
    s = Spectrum([1, 2, 3, 4, 5], [2, 2, 2, 2, 2])
    assert np.array_equal(s.bin([2, 3], preserve_power=False), [1, 3])


def test_spectrum_integrate_no_bounds():
    s = Spectrum([1, 2, 3, 4, 5], [2, 2, 2, 2, 2])
    assert s.integrate() == 8.0


def test_spectrum_integrate_bounds():
    s = Spectrum([1, 2, 3, 4, 5], [2, 2, 2, 2, 2])
    assert s.integrate(2, 4) == 4.0


def test_spectrum_ends():
    s = Spectrum([1, 2, 3, 4, 5], [0.1, 1, 100, 1, 0.1])
    assert np.array_equal(s.ends(tol=0.001), [1, 3])


def test_spectrum_trim():
    s = Spectrum([1, 2, 3, 4, 5], [0.1, 1, 100, 1, 0.1])
    s.trim(tol=0.001)
    assert np.array_equal(s.wave, [2, 3, 4])
    assert np.array_equal(s.value, [1, 100, 1])


def test_spectrum_trim_value_all_zeros():
    s = Spectrum([1, 2, 3, 4, 5], [0, 0, 0, 0, 0])
    s.trim(tol=0.0001)
    assert np.array_equal(s.wave, [1, 2, 3, 4, 5])
    assert np.array_equal(s.value, [0, 0, 0, 0, 0])


def test_spectrum_crop():
    s = Spectrum([1, 2, 3, 4, 5], [0.1, 1, 100, 1, 0.1])
    s.crop(2, 5)
    assert np.array_equal(s.wave, [2, 3, 4, 5])
    assert np.array_equal(s.value, [1, 100, 1, 0.1])


def test_spectrum_crop_outside_wave_limits():
    s = Spectrum([1, 2, 3, 4, 5], [0.1, 1, 100, 1, 0.1])
    s.crop(2, 10)
    assert np.array_equal(s.wave, [2, 3, 4, 5])
    assert np.array_equal(s.value, [1, 100, 1, 0.1])



# Lentil
# Heart-healthy physical optics
#
# Copyright (c) 2020-2024, California Institute of Technology ("Caltech"). 
# U.S. Government sponsorship acknowledged.

__version__ = '0.8.4'

from lentil.ptype import ptype

none = ptype('none')
pupil = ptype('pupil')
image = ptype('image')
tilt = ptype('tilt')
transform = ptype('transform')

from lentil.convolvable import jitter, smear

from lentil import detector

from lentil import fourier

from lentil.plane import (
	Plane,
	Pupil,
	Image,
    Tilt,
	DispersiveTilt,
	Grism,
	LensletArray,
	Rotate,
	Flip
)

from lentil.propagate import (
    propagate_dft, 
    propagate_fft,
    scratch_shape
)

from lentil import radiometry

from lentil.segmented import hex_segments

from lentil.shape import (
    circle,
	hexagon,
	rectangle,
    spider
)

from lentil import util
from lentil.util import (
	centroid,
	pad,
    subarray,
	window,
	boundary,
	rebin,
	rescale,
	pixelscale_nyquist,
	min_sampling,
	normalize_power,
	sanitize_shape,
	sanitize_bandpass
)

from lentil.wavefront import Wavefront

from lentil.wfe import power_spectrum, translation_defocus

from lentil.zernike import (
	zernike,
	zernike_compose,
	zernike_fit,
	zernike_remove,
	zernike_basis,
	zernike_coordinates
)



import pytest
import numpy as np

import lentil


def test_boundary():
    mask = lentil.hexagon((256, 256), 100, rotate=True)
    bounds = lentil.boundary(mask)
    assert np.array_equal(bounds, [28, 228, 41, 215])


def test_rebin():
    img = np.array([[1, 1, 2, 2], [1, 1, 2, 2], [3, 3, 4, 4], [3, 3, 4, 4]])
    factor = 2
    img_rebinned = lentil.util.rebin(img, factor)
    assert np.array_equal(img_rebinned, np.array([[4, 8], [12, 16]]))


def test_rebin_cube():
    img = np.zeros((2, 4, 4))
    img[0] = np.array([[1, 1, 2, 2], [1, 1, 2, 2], [3, 3, 4, 4], [3, 3, 4, 4]])
    img[1] = np.array([[2, 2, 4, 4], [2, 2, 4, 4], [6, 6, 8, 8], [6, 6, 8, 8]])
    factor = 2
    img_rebinned = lentil.util.rebin(img, factor)
    img_rebinned_expected = np.zeros((2, 2, 2))
    img_rebinned_expected[0] = np.array([[4, 8], [12, 16]])
    img_rebinned_expected[1] = np.array([[8, 16], [24, 32]])
    assert np.array_equal(img_rebinned, img_rebinned_expected)


def test_rescale_unitary():
    a = np.random.uniform(low=0, high=1, size=(100, 100))
    b = lentil.util.rescale(a, scale=0.5, order=3, mode='nearest', unitary=True)
    assert np.isclose(np.sum(a), np.sum(b))


def test_pad():
    img = np.ones((3, 3))
    out = lentil.util.pad(img, (5, 5))
    truth = np.zeros((5, 5))
    truth[1:4, 1:4] = 1

    assert np.array_equal(out, truth)


def test_pad_cube():
    img = np.ones((2, 3, 3))
    out = lentil.util.pad(img, (5, 5))
    truth = np.zeros((2, 5, 5))
    truth[:, 1:4, 1:4] = 1

    assert np.array_equal(out, truth)


def test_pad_shrink():
    img = np.random.uniform(size=(5, 5))
    out = lentil.util.pad(img, (3, 3))
    assert np.array_equal(out, img[1:4, 1:4])


win_data = np.diag(np.arange(1, 11))
for i in np.diag(win_data):
    win_data[i-1, i-1:] = win_data[i-1, i-1]
    win_data[i-1:, i-1] = win_data[i-1, i-1]


def test_window_none():
    assert np.array_equal(lentil.util.window(win_data, shape=None, slice=None), win_data)


def test_window_single():
    assert np.array_equal(lentil.util.window(5), np.array(5))


def test_even_window_shape():
    assert np.array_equal(lentil.util.window(win_data, shape=(2, 2)), np.array([[5, 5], [5, 6]]))


def test_even_window_window():
    assert np.array_equal(lentil.util.window(win_data, slice=(0, 2, 0, 2)), np.array([[1, 1], [1, 2]]))


def test_odd_window_shape():
    assert np.array_equal(lentil.util.window(win_data[:-1, :-1], shape=(2, 2)), np.array([[4, 4], [4, 5]]))


def test_odd_window_window():
    assert np.array_equal(lentil.util.window(win_data[:-1, :-1], slice=(0, 2, 0, 2)), np.array([[1, 1], [1, 2]]))


def test_window_shape_window():
    assert np.array_equal(lentil.util.window(win_data, shape=(2, 2), slice=(8, 10, 8, 10)), np.array([[9, 9], [9, 10]]))


def test_window_shape_mismatch():
    with pytest.raises(AssertionError):
        lentil.util.window(win_data, shape=(3, 3), slice=(0, 2, 0, 2))


def test_pixelscle_nyquist():
    wave = np.random.uniform(low=350e-9, high=750e-9)
    f_number = np.random.uniform(low=5, high=30)
    assert lentil.util.pixelscale_nyquist(wave, f_number) == wave * f_number / 2


amp = lentil.circle((512, 512), 256)


def test_amp_norm():
    # make sure we're not starting with an already normalized array!
    assert np.sum(np.abs(amp)**2) != 1.0


def test_normalize_power():
    amp_norm = lentil.normalize_power(amp)
    assert np.isclose(np.sum(np.abs(amp_norm)**2), 1.0)


def test_normalize_power_2():
    amp_norm = lentil.normalize_power(amp, power=2)
    assert np.isclose(np.sum(np.abs(amp_norm)**2), 2.0)



import math
import numpy as np

import pytest
import lentil


size = (10, 10)


class RandomPlane(lentil.Plane):
    def __init__(self):
        super().__init__(pixelscale=1,
                         amplitude=np.random.uniform(size=size),
                         opd=np.random.uniform(size=size))


def test_default_plane():
    # Ensure that a default Plane creates an object that won't have any
    # impact on an optical system (a perfect optic with no wavefront error 
    # and perfect optical and spectral transmission).

    p = lentil.Plane()
    assert p.pixelscale == None
    assert np.all(p.amplitude == 1)
    assert np.all(p.opd == 0)
    assert p.mask == p.amplitude


def test_amp_alias():
    p = lentil.Plane(amp=10)
    assert p.amplitude == 10


def test_amp_alias_error():
    with pytest.raises(TypeError):
        lentil.Plane(amplitude=10, amp=10)


class PlaneAttributeOverload(lentil.Plane):
    def __init__(self):
        super().__init__()

    @property
    def amplitude(self):
        return np.array(1)
    
    @property
    def opd(self):
        return np.array(2)
    

def test_plane_overload_propertyes():
    p = PlaneAttributeOverload()
    
    assert p.amplitude == 1
    assert p.opd == 2
    assert p.mask == 1

    with pytest.raises(AttributeError):
        p.opd = 5

    with pytest.raises(AttributeError):
        p.amplitude = 5


def test_plane_fit_tilt_inplace():
    p = RandomPlane()
    p_copy = p.fit_tilt(inplace=False)
    p_inplace = p.fit_tilt(inplace=True)

    assert p_copy is not p
    assert p_inplace is p


def test_wavefront_plane_multiply():
    p = RandomPlane()
    w = lentil.Wavefront(650e-9)

    w1 = p.multiply(w)

    slc = lentil.helper.boundary_slice(p.mask)
    phasor = p.amplitude[slc] * np.exp(2*np.pi*1j*p.opd[slc]/w.wavelength)

    assert np.array_equal(w1.data[0].data, phasor)


def test_wavefront_plane_multiply_overlapping_segment_slices():
    seg = lentil.hexagon((64, 64), 32, antialias=False)
    seg = seg[5:60, :]

    segmask = np.zeros((2, 128, 128))
    segmask[0, 0:55, 2:66] = seg
    segmask[1, 29:84, 55:119] = seg
    mask = np.sum(segmask, axis=0)

    pupil = lentil.Pupil(amplitude=mask, mask=segmask, pixelscale=1 / 256, focal_length=10)

    w = lentil.Wavefront(500e-9)
    w *= pupil

    assert np.array_equal(mask, w.intensity)


class CircularPupil(lentil.Pupil):
    def __init__(self):
        super().__init__(focal_length=10,
                         pixelscale=2/256,
                         amplitude=lentil.circle((256, 256), 128),
                         opd=np.zeros((256, 256)))


def test_wavefront_pupil_multiply():
    p = CircularPupil()
    w = lentil.wavefront.Wavefront(650e-9)
    w = p.multiply(w)
    phasor = p.amplitude * np.exp(1j*p.opd * 2 * np.pi / w.wavelength)

    assert np.array_equal(w.data[0].data, phasor)
    assert w.focal_length == p.focal_length


def test_pupil_rescale_power():
    p = CircularPupil()
    pr = p.rescale(3)

    amp_power = np.sum(np.abs(p.amplitude)**2)
    ampr_power = np.sum(np.abs(pr.amplitude)**2)
    assert math.isclose(amp_power, ampr_power, rel_tol=1e-3)


def test_dispersive_tilt_center():
    dispersion = [1, 650e-9]
    trace = [2, 0]
    dt = lentil.DispersiveTilt(dispersion=dispersion, trace=trace)

    x0 = np.random.uniform(low=-5, high=5)
    y0 = np.random.uniform(low=-5, high=5)
    x, y = dt.shift(wavelength=650e-9, xs=x0, ys=y0)
    assert np.all((x == x0, y == y0))


def test_dispersive_tilt_shift():
    dt = lentil.DispersiveTilt(trace=[1,1], dispersion=[1,650e-9])
    wave = 900e-9
    x, y = dt.shift(wavelength=wave)

    assert x == (wave - dt.dispersion[1])/np.sqrt(2)
    assert y == 1+x




import numpy as np
import lentil


def test_jitter_scale():
    a = np.zeros((11, 11))
    a[5, 5] = 1

    scale = 2
    b = lentil.jitter(a, scale)

    def fwhm(arr):
        arr = arr / np.max(arr)
        arr[arr < 0.5] = 0
        extents = lentil.util.boundary(arr)
        return np.max([extents[1] - extents[0], extents[3] - extents[2]])

    assert fwhm(b) // 2 == scale


def test_jitter_unitary():
    a = np.zeros((11, 11))
    a[5, 5] = 1

    scale = 2
    b = lentil.jitter(a, scale)
    assert np.isclose(np.sum(a), np.sum(b))


def test_smear_unitary():
    a = np.zeros((11, 11))
    a[5, 5] = 1

    distance = 2
    b = lentil.smear(a, distance)
    assert np.isclose(np.sum(a), np.sum(b))





import math
import numpy as np
import scipy.interpolate
import scipy.special
import lentil


def test_amplitude_normalize_power(pupil):
    p = pupil(focal_length=10, diameter=1, shape=(256,256), radius=120, coeffs=None)
    w = lentil.Wavefront(wavelength=650e-9)
    w *= p
    w = lentil.propagate_dft(w, shape=(64,64), pixelscale=5e-6, oversample=1)
    psf = w.intensity
    assert (np.sum(psf) <= 1) and (np.sum(psf) >= 0.95)


def test_amplitude_normalize_power_oversample(pupil):
    p = pupil(focal_length=10, diameter=1, shape=(256,256), radius=120, coeffs=None)
    w = lentil.Wavefront(wavelength=650e-9)
    w *= p
    w = lentil.propagate_dft(w, shape=(64,64), pixelscale=5e-6, oversample=2)
    psf = w.intensity
    assert (np.sum(psf) <= 1) and (np.sum(psf) >= 0.95)


def test_propagate_airy(airy2, pupil):
    # we set oversample = 1 and apply an "oversampling" factor of 25 to pixelscale
    # so that we can more easily enforce an odd npix. having an odd npix enables
    # comparison between numerical propagations via dft2() and the airy2() function
    # since their DC pixels will be coincident.
    p = pupil(focal_length=10, diameter=1, shape=(512,512), radius=250, coeffs=None)

    psf_airy = airy2(diameter=p.diameter, focal_length=p.focal_length, wavelength=650e-9,
                     pixelscale=[5e-6, 5e-6], shape=(511, 511), oversample=1)
    psf_airy = psf_airy/np.max(psf_airy)

    w = lentil.Wavefront(wavelength=650e-9)
    w *= p
    w = lentil.propagate_dft(w, shape=511, pixelscale=5e-6, oversample=1)
    psf = w.intensity
    psf = psf/np.max(psf)
    assert np.all(np.isclose(psf, psf_airy, atol=1e-3))


def test_propagate_fit_tilt(pupil):
    coeffs = 5e-6 * np.random.uniform(low=-0.5, high=0.5, size=3)
    coeffs[0] = 0
    p = pupil(focal_length=10, diameter=1, shape=(256,256), radius=120, coeffs=coeffs)

    w = lentil.Wavefront(650e-9)
    w = p.multiply(w)
    w = lentil.propagate_dft(w, shape=128, pixelscale=5e-6, oversample=2)
    psf = w.intensity

    p.fit_tilt(inplace=True)
    w_fit_tilt = lentil.Wavefront(650e-9)
    w_fit_tilt = w_fit_tilt * p
    w_fit_tilt = lentil.propagate_dft(w_fit_tilt, shape=128, pixelscale=5e-6, oversample=2)
    psf_fit_tilt = w_fit_tilt.intensity

    # threshold the PSFs so that the centroiding is consistent
    psf[psf < 1e-5] = 0
    psf_fit_tilt[psf_fit_tilt < 1e-5] = 0

    delta = np.abs(np.asarray(lentil.util.centroid(psf)) - np.asarray(lentil.util.centroid(psf_fit_tilt)))
    assert np.all(delta <= 1e-10)


def test_propagate_tilt_analytic(pupil):
    oversample = 10
    pixelscale = 5e-6
    npix = np.array([64, 64])
    
    coeffs = 5e-6 * np.random.uniform(low=-0.5, high=0.5, size=3)
    coeffs[0] = 0
    p = pupil(focal_length=10, diameter=1, shape=(256,256), radius=120, coeffs=coeffs)

    w = lentil.Wavefront(650e-9)
    w = p.multiply(w)
    w = lentil.propagate_dft(w, shape=npix, pixelscale=pixelscale, oversample=oversample)
    psf = w.intensity

    psf = psf/np.max(psf)
    psf[psf < 0.2] = 0  # threshold for centroiding

    center = npix//2 * oversample
    centroid = np.asarray(lentil.util.centroid(psf))
    shift = centroid - center

    pupil_tilt = coeffs[1:3]

    # The factor of 4 gets you from RMS to PV
    analytic_shift = ((pupil_tilt/p.diameter)*p.focal_length/pixelscale) * oversample * 4

    # Analytic shift is in terms of (x, y) but we really want (r, c) to match the centroid
    analytic_shift = analytic_shift[1], -analytic_shift[0]

    assert np.all((np.abs(shift - analytic_shift)/oversample) < 0.2)


def test_propagate_tilt_fit_tilt_analytic(pupil):
    oversample = 10
    pixelscale = 5e-6
    npix = np.array([64, 64])
    
    coeffs = 5e-6 * np.random.uniform(low=-0.5, high=0.5, size=3)
    coeffs[0] = 0
    p = pupil(focal_length=10, diameter=1, shape=(256,256), radius=120, coeffs=coeffs)

    p.fit_tilt(inplace=True)
    w = lentil.Wavefront(650e-9)
    w = w * p
    w = lentil.propagate_dft(w, shape=npix, pixelscale=pixelscale, oversample=oversample)
    psf = w.intensity

    psf = psf/np.max(psf)
    psf[psf < 0.2] = 0  # threshold for centroiding

    center = npix//2 * oversample
    centroid = np.asarray(lentil.util.centroid(psf))
    shift = centroid - center

    pupil_tilt = coeffs[1:3]

    # The factor of 4 gets you from RMS to PV
    analytic_shift = ((pupil_tilt/p.diameter)*p.focal_length/pixelscale) * oversample * 4

    # Analytic shift is in terms of (x, y) but we really want (r, c) to match the centroid
    analytic_shift = analytic_shift[1], -analytic_shift[0]

    assert np.all((np.abs(shift - analytic_shift)/oversample) < 0.2)


def test_propagate_no_output(pupil):

    p = pupil(focal_length=10, diameter=1, shape=(255,255), radius=120, coeffs=[0, 1e-3])
    p.fit_tilt(inplace=True)
    w = lentil.Wavefront(650e-9)
    w *= p
    w = lentil.propagate_dft(w, shape=(64,64), pixelscale=5e-6, oversample=2)
    assert np.all(w.intensity == 0)


def test_propagate_resample(pupil):

    coeffs = np.random.uniform(low=-200e-9, high=200e-9, size=6)

    p = pupil(focal_length=10, diameter=1, shape=(256,256), radius=120, coeffs=coeffs)
    w = lentil.Wavefront(650e-9)
    w *= p
    wi = lentil.propagate_dft(w, shape=(64,64), pixelscale=5e-6, oversample=10)

    p2 = p.rescale(scale=3)
    w2 = lentil.Wavefront(650e-9)
    w2 *= p2
    w2i = lentil.propagate_dft(w2, shape=(64,64), pixelscale=5e-6, oversample=10)

    # compute cross correlation between wi and w2i
    xc = np.fft.ifftshift(np.conj(np.fft.fft2(wi.intensity)) * np.fft.fft2(w2i.intensity))
    cent = lentil.centroid(np.abs(xc))

    assert np.allclose(cent, [320, 320])
    assert math.isclose(np.sum(wi.intensity), np.sum(w2i.intensity), rel_tol=1e-2)



import pytest
import numpy as np
import lentil


def test_collect_charge_scalar():
    img = np.random.uniform(low=0, high=100, size=(10, 10))
    qe = np.random.uniform()
    out = lentil.detector.collect_charge(img, 1, qe)
    assert np.array_equal(out, img*qe)


def test_collect_charge_array():
    img = np.random.uniform(low=0, high=100, size=(5, 10, 10))
    qe = np.random.uniform(size=5)
    out = lentil.detector.collect_charge(img, np.ones(5), qe)
    assert np.array_equal(out, np.einsum('ijk,i->jk', img, qe))


def test_collect_charge_spectrum():
    img = np.random.uniform(low=0, high=100, size=(5, 10, 10))
    wave = np.arange(1, 6)
    value = np.random.uniform(size=5)
    qe = lentil.radiometry.Spectrum(wave, value)
    out = lentil.detector.collect_charge(img, wave, qe)
    assert np.array_equal(out, np.einsum('ijk,i->jk', img, value))


def test_collect_charge_bayer_even():
    img = np.random.uniform(low=0, high=100, size=(5, 2, 2))
    qe_red = np.random.uniform(size=5)
    qe_green = np.random.uniform(size=5)
    qe_blue = np.random.uniform(size=5)
    bayer_pattern = 'RGGB'
    out = lentil.detector.collect_charge_bayer(img, np.ones(5), qe_red,
                                               qe_green, qe_blue, bayer_pattern)

    ul = np.sum(img[:, 0, 0]*qe_red)
    ur = np.sum(img[:, 0, 1]*qe_green)
    ll = np.sum(img[:, 1, 0]*qe_green)
    lr = np.sum(img[:, 1, 1]*qe_blue)

    assert np.allclose(out, np.array([[ul, ur], [ll, lr]]))


def collect_charge_bayer_odd():
    img = np.random.uniform(low=0, high=100, size=(5, 3, 3))
    qe_red = np.random.uniform(size=5)
    qe_green = np.random.uniform(size=5)
    qe_blue = np.random.uniform(size=5)
    bayer_pattern = 'RGGB'
    out = lentil.detector.collect_charge_bayer(img, np.ones(5), qe_red,
                                               qe_green, qe_blue, bayer_pattern)

    a = np.sum(img[:, 0, 0]*qe_red)
    b = np.sum(img[:, 0, 1]*qe_green)
    c = np.sum(img[:, 0, 2]*qe_red)
    d = np.sum(img[:, 1, 0]*qe_green)
    e = np.sum(img[:, 1, 1]*qe_blue)
    f = np.sum(img[:, 1, 2]*qe_green)
    g = np.sum(img[:, 2, 0]*qe_red)
    h = np.sum(img[:, 2, 1]*qe_green)
    i = np.sum(img[:, 2, 2]*qe_red)

    assert np.allclose(out, np.array([[a, b, c], [d, e, f], [g, h, i]]))


def test_collect_charge_bayer_oversample():
    img = np.random.uniform(low=0, high=100, size=(5, 4, 4))
    qe_red = np.random.uniform(size=5)
    qe_green = np.random.uniform(size=5)
    qe_blue = np.random.uniform(size=5)
    bayer_pattern = 'RGGB'
    out = lentil.detector.collect_charge_bayer(img, np.ones(5), qe_red,
                                               qe_green, qe_blue, bayer_pattern,
                                               oversample=2)

    ul = np.einsum('ijk,i->jk', img[:, 0:2, 0:2], qe_red)
    ur = np.einsum('ijk,i->jk', img[:, 0:2, 2:4], qe_green)
    ll = np.einsum('ijk,i->jk', img[:, 2:4, 0:2], qe_green)
    lr = np.einsum('ijk,i->jk', img[:, 2:4, 2:4], qe_blue)

    assert np.allclose(out, np.block([[ul, ur], [ll, lr]]))


def test_pixel_unitary():
    a = np.random.uniform(low=0, high=1, size=(100, 100))
    b = lentil.detector.pixel(a, oversample=1)
    assert np.isclose(np.sum(a), np.sum(b))


def test_pixelate():
    # we've already tested rescale and the pixel MTF separately, so we're just
    # going to ensure the returned image has the right shape here
    img = np.ones((10, 10))
    out = lentil.detector.pixelate(img, 2)
    assert out.shape == (5, 5)


def test_adc_saturation_capacity():
    img = 10 * np.ones((10, 10))
    out = lentil.detector.adc(img, gain=1, saturation_capacity=1)
    assert np.array_equal(out, np.ones((10, 10)))


def test_adc_warn_saturate():
    with pytest.warns(UserWarning):
        img = 10 * np.ones((10, 10))
        out = lentil.detector.adc(img, gain=1, saturation_capacity=1,
                                  warn_saturate=True)


def test_adc_float_gain():
    g = np.random.uniform(size=1)
    img = 10*np.random.uniform(size=(10, 10))
    out = lentil.detector.adc(img, g, saturation_capacity=None)
    assert np.array_equal(out, np.floor(g*img))


def test_adc_array_gain():
    g = np.random.uniform(size=(10, 10))
    img = 10*np.random.uniform(size=(10, 10))
    out = lentil.detector.adc(img, g, saturation_capacity=None)
    assert np.array_equal(out, np.floor(g*img))


def test_adc_nonlinear_gain():
    g = np.random.uniform(low=0.9, high=1.5, size=4)
    img = 10*np.ones((10, 10))
    out = lentil.detector.adc(img, g, saturation_capacity=None)
    assert np.isclose(out[0, 0], np.floor(np.polyval(np.append(g, 0), 10)))


def test_adc_nonlinear_gain_cube():
    g = np.random.uniform(low=0.9, high=1.5, size=(4, 10, 10))
    img = 10*np.ones((10, 10))
    out = lentil.detector.adc(img, g, saturation_capacity=None)
    index = np.random.uniform(low=0, high=9, size=2)
    r = int(index[0])
    c = int(index[1])
    assert np.isclose(out[r, c], np.floor(np.polyval(np.append(g[:, r, c], 0), 10)))


def test_adc_invalid_gain():
    img = 10*np.ones((10, 10))
    with pytest.raises(ValueError):
        out = lentil.detector.adc(img, gain=np.ones((1, 1, 1, 1)))


def test_adc_dtype():
    g = np.random.uniform(size=1)
    img = 10*np.random.uniform(size=(10, 10))
    out = lentil.detector.adc(img, g, dtype=np.int8)
    assert out.dtype == np.int8


def test_shot_noise_poisson():
    img = np.random.uniform(low=0, high=1e5, size=(2, 2))
    shot1 = lentil.detector.shot_noise(img, method='poisson')
    shot2 = lentil.detector.shot_noise(img, method='poisson')
    assert np.all(shot1 != shot2)
    assert shot1.shape == (2, 2)


def test_shot_noise_poisson_seed():
    img = np.random.uniform(low=0, high=1e5, size=(2, 2))
    shot1 = lentil.detector.shot_noise(img, method='poisson', seed=12345)
    shot2 = lentil.detector.shot_noise(img, method='poisson', seed=12345)
    assert np.array_equal(shot1, shot2)


def test_shot_noise_poisson_negative():
    with pytest.raises(ValueError):
        lentil.detector.shot_noise(-10, method='poisson')


def test_shot_noise_poisson_big():
    with pytest.raises(ValueError):
        lentil.detector.shot_noise(10e100, method='poisson')


def test_shot_noise_gaussian():
    img = np.random.uniform(low=0, high=1e5, size=(2, 2))
    shot1 = lentil.detector.shot_noise(img, method='gaussian')
    shot2 = lentil.detector.shot_noise(img, method='gaussian')
    assert np.all(shot1 != shot2)
    assert shot1.shape == (2, 2)


def test_shot_noise_gaussian_seed():
    img = np.random.uniform(low=0, high=1e5, size=(2, 2))
    shot1 = lentil.detector.shot_noise(img, method='gaussian', seed=12345)
    shot2 = lentil.detector.shot_noise(img, method='gaussian', seed=12345)
    assert np.array_equal(shot1, shot2)


@pytest.mark.filterwarnings("ignore:invalid value encountered in sqrt")
def test_shot_noise_gaussian_negative():
    with pytest.raises(ValueError):
        lentil.detector.shot_noise(-10, method='gaussian')


def test_read_noise_gaussian():
    img = np.random.uniform(low=0, high=1e5, size=(2, 2))
    read1 = lentil.detector.read_noise(img, electrons=100)
    read2 = lentil.detector.read_noise(img, electrons=100)
    assert np.all(read1 != read2)
    assert read1.shape == (2, 2)


def test_read_noise_gaussian_seed():
    img = np.random.uniform(low=0, high=1e5, size=(2, 2))
    read1 = lentil.detector.read_noise(img, electrons=100, seed=12345)
    read2 = lentil.detector.read_noise(img, electrons=100, seed=12345)
    assert np.array_equal(read1, read2)


def test_charge_diffusion_shape():
    img = np.random.uniform(low=0, high=1e5, size=(10, 10))
    out = lentil.detector.charge_diffusion(img, sigma=0.5, oversample=1)
    assert out.shape == img.shape


def test_charge_diffusion_shape_oversample():
    img = np.random.uniform(low=0, high=1e5, size=(10, 10))
    out = lentil.detector.charge_diffusion(img, sigma=0.5, oversample=2)
    assert out.shape == img.shape


def test_dark_current_fpn():
    dark1 = lentil.detector.dark_current(rate=100, shape=(10, 10), fpn_factor=0.1, seed=12345)
    dark2 = lentil.detector.dark_current(rate=100, shape=(10, 10), fpn_factor=0.1, seed=12345)
    assert np.array_equal(dark1, dark2)


def test_rule07_dark_current_shape():
    out = lentil.detector.rule07_dark_current(temperature=25,
                                              cutoff_wavelength=6e-6,
                                              pixelscale=18e-6,
                                              shape=(2, 2))

    assert out.shape == (2, 2)


def test_rule07_dark_current_low_cutoff():
    out = lentil.detector.rule07_dark_current(temperature=110,
                                              cutoff_wavelength=4e-6,
                                              pixelscale=18e-6,
                                              shape=1)

    assert np.isclose(out, 65)


def test_rule07_dark_current_high_cutoff():
    out = lentil.detector.rule07_dark_current(temperature=110,
                                              cutoff_wavelength=5e-6,
                                              pixelscale=18e-6,
                                              shape=1)

    assert np.isclose(out, 10403)


def test_cosmic_rays():
    # we're going to punt on this one for now
    frame = lentil.detector.cosmic_rays((10, 10), 0.1*np.ones(3), 1, rate=1)
    assert frame.shape == (10, 10)



import collections

import numpy as np

import lentil


def hex_segments(rings, seg_radius, seg_gap, rotate=False, antialias=True,
                 flatten=False, pad=2, drop=(0,)):
    """Draw a segmented aperture made up of hexagonal segments

    Parameters
    ----------
    rings : int
        Number of segment rings. Must be >= 1.
    seg_radius : float
        Segment outscribing radius in pixels
    seg_gap : float
        Spacing between adjacent segments in pixels
    rotate : bool, optional
        Rotate segments so that flat sides are aligned with the Y direction 
        instead of the default orientation which is aligned with the X 
        direction.
    antialias : bool, optional
        If True (default), the segment edges are antialiased.
    flatten : bool, optional
        If True, the individual segment masks are flattened into a single
        global mask. Default is False.
    pad : int, optional
        Number of additional pixels to zero-pad output by. Default is 2.
    drop : list_like, optional
        Segments to exclude from output. 0 represents the central segment. 
    
    Returns
    -------
    ndarray

    Examples
    --------
    .. plot::
        :include-source:
        :context: reset
        :scale: 50

        >>> import matplotlib.pyplot as plt
        >>> import lentil
        >>> mask = lentil.hex_segments(rings=2, seg_radius=64, seg_gap=2)
        >>> fig, ax = plt.subplots(nrows=1, ncols=2, figsize=(5, 2))
        >>> ax[0].imshow(np.sum(mask, axis=0))
        >>> ax[0].set_title('Full aperture')
        >>> ax[1].imshow(mask[0])
        >>> ax[1].set_title('Segment 1')

    References
    ----------
    [1] Hexagonal Grids (https://www.redblobgames.com/grids/hexagons/)

    """
    inner_radius = seg_radius * np.sqrt(3)/2
    size = np.ceil((rings * 2 + 1) * inner_radius * 2 + (rings * 2) * seg_gap + pad * 2).astype(int)
    shape = np.broadcast_to(size, (2,))

    mask = []

    if 0 not in drop:
        mask.append(lentil.hexagon(shape, seg_radius, shift=(0, 0), antialias=antialias, rotate=rotate))

    seg = 1
    for ring in range(1, rings+1):
        for h in hex_ring(ring):
            r, c = hex_to_rc(h, seg_radius + seg_gap/2, rotate)
            if seg not in drop:
                mask.append(lentil.hexagon(shape, seg_radius, shift=(r, c), antialias=antialias, rotate=rotate))
            seg += 1
    
    mask = np.asarray(mask)
    if flatten:
        mask = np.sum(mask, axis=0)
    
    return mask


# the functions below for working with hex grids was copied almost
# exactly from http://www.redblobgames.com/grids/hexagons/

_Hex = collections.namedtuple("Hex", ["q", "r", "s"])
def Hex(q, r, s):
    assert not (round(q + r + s) != 0), "q + r + s must be 0"
    return _Hex(q, r, s)


def hex_add(a, b):
    return Hex(a.q + b.q, a.r + b.r, a.s + b.s)


hex_directions = [Hex(1, 0, -1), Hex(1, -1, 0), Hex(0, -1, 1), 
                  Hex(-1, 0, 1), Hex(-1, 1, 0), Hex(0, 1, -1)]


def hex_direction(direction):
    return hex_directions[direction]


def hex_neighbor(hex, direction):
    return hex_add(hex, hex_direction(direction))


def hex_ring(radius):
    results = []
    hex = Hex(-radius, radius, 0)

    for i in range(6):
        for j in range(radius):
            results.append(hex)
            hex = hex_neighbor(hex, i)
    return results


def hex_to_xy(hex, radius, rotate=False):
    if rotate:
        x = radius * (np.sqrt(3) * hex.q + np.sqrt(3)/2 * hex.r)
        y = radius * (3/2 * hex.r)
    else:
        x = radius * (3/2 * hex.q)
        y = radius * (np.sqrt(3)/2 * hex.q + np.sqrt(3) * hex.r)
        
    return x, y


def hex_to_rc(hex, radius, rotate=False):
    x, y = hex_to_xy(hex, radius, rotate)
    return -y, x





PTYPES = ('none', 'pupil', 'image', 'tilt', 'transform')


def ptype(ptype):
    """Create a plane type object

    Parameters
    ----------
    ptype
        Object to be converted to a plane type object. Can be
        a string, a ptype, or None.
    """
    if ptype is None:
        ptype = 'none'

    if isinstance(ptype, PType):
        return ptype
    else:
        return PType(ptype)


class PType:
    """An object to hold the type of a :class:`~lentil.Plane` or 
    :class:`~lentil.Wavefront`

    Parameters
    ----------
    ptype : str
        Plane type
    """
    def __init__(self, ptype):
        if ptype not in PTYPES:
            raise TypeError(f"plane type '{ptype}' not understood")
        self._key = ptype

    def __eq__(self, other):
        if self._key == other._key:
            return True
        else:
            return False
        
    def __hash__(self):
        return hash(self._key)
    
    def __repr__(self):
        return f"ptype('{self._key}')"

    def __str__(self):
        return self._key
    


import sys
from itertools import combinations
import numpy as np

import lentil.extent

class Field:
    """
    Two-dimensional discretely sampled complex field.

    Parameters
    ----------
    data : array_like
        Array containing the sampled field. Note that real-valued data is
        cast to complex.
    pixelscale : float or None, optional
        Data array spatial sampling. If None (default), the field is assumed to
        be broadcastable to any legal shape without interpolation.
    offset : (2,) array_like of ints or None, optional.
        Shift of the Field in (row, column) from (0, 0). If None (default),
        offset = (0, 0).
    tilt : list, optional
        List of objects which implement a ``shift`` method. This method should
        accept the following parameters:

        ``shift(xs, ys, z, wavelength)``

        and return an updated x and y shift. If None (default), tilt = [].

    """
    __slots__ = ('data', 'offset', 'tilt', 'pixelscale', 'extent')

    def __init__(self, data, pixelscale=None, offset=None, tilt=None):
        #: ndarray : Complex field data
        self.data = np.asarray(data, dtype=complex) 
        
        self.pixelscale = pixelscale
        """Spatial sampling of data
        
        If None (default), the Field is assumed to be broadcastable to any
        legal shape without interpolation.

        Returns
        -------
        tuple of ints or None
        """
        
        #: tuple of ints : Field offset from (0, 0).
        self.offset = offset if offset is not None else [0, 0]
        
        #: list : List of objects that implement the tilt interface defined
        # in :class:`~lentil.plane.TiltInterface`
        self.tilt = tilt if tilt else []

        #: tuple of ints : Extent of ``data``
        self.extent = lentil.extent.array_extent(self.shape, self.offset)

    @property
    def shape(self):
        """
        Tuple of Field dimensions

        Returns
        -------
        tuple of ints
        """
        return self.data.shape

    @property
    def size(self):
        """
        Number of elements in the Field

        Returns
        -------
        int
        """
        return self.data.size

    def __mul__(self, other):
        """
        Multiply fields element-wise.

        Parameters
        ----------
        x1, x2 : :class:`~lentil.field.Field`
            Fields to be multiplied. If ``x1.shape != x2.shape``, they must
            be broadcastable to a common shape (which becomes the shape of the
            output).

        Returns
        -------
        y : :class:`~lentil.field.Field`
            The product of ``x1`` and ``x2``, appropriately sized and offset.

        Notes
        -----
        The output Field data and offset are computed according to the following
        rules:

            * If both operands are scalars, the result is 0 unless the operands
            share the same offset.
            * A single scalar operand is broadcast to the larger operand's shape and
            inherits the larger operand's offset. This enables default planar
            fields (``data = 1+0j``, offset = [0, 0]) to automatically grow in
            size when multiplied by a "real" field.
        """
        tilt = self.tilt + other.tilt

        if self.size == 1 and other.size == 1:
            data, offset = self._mul_scalar(other)
        else:
            # Note that _mul_array is optimized to also handle scalar * array
            data, offset = self._mul_array(other)

        return Field(data=data, offset=offset, tilt=tilt)
    
    def _mul_scalar(self, other):
        """
        Multiply two Fields when both are scalars
        """
        if np.array_equal(self.offset, other.offset):
            data = self.data * other.data
            offset = self.offset
        else:
            data = []
            offset = None
        return data, offset
    
    def _mul_array(self, other):
        """
        Multiply two fields when at least one is an array
        """
        self_data, self_offset, other_data, other_offset = _mul_broadcast(
            self.data, self.offset, other.data, other.offset
        )
        self_extent = lentil.extent.array_extent(self_data.shape, self_offset)
        other_extent = lentil.extent.array_extent(other_data.shape, other_offset)

        if lentil.extent.intersect(self_extent, other_extent):
            self_slice, other_slice = lentil.extent.intersection_slices(self_extent, other_extent)
            data = self_data[self_slice] * other_data[other_slice]
            offset = lentil.extent.intersection_shift(self_extent, other_extent)
        else:
            data = []
            offset = None
        return data, offset

    def shift(self, z, wavelength, pixelscale, oversample, indexing='ij'):
        """
        Compute Field shift due to associated :class:`~lentil.Tilt` objects

        Parameters
        ----------
        z : float
            Propagation distance
        wavelength : float
            Wavelength
        pixelscale : (2,) float
            Output plane spatial sampling
        oversample : int
            Output plane oversampling factor
        indexing : {'ij', 'xy'}, optional
            Cartesian ('xy') or matrix ('ij', default) indexing of output.

        Returns
        -------
        shift : tuple of floats
            Shifted location of the center of the field

        Notes
        -----
        This function is capable of doing the conversion between the (x, y) coordinates
        of a Tilt and the (r, c) coordinates of a Field.

        """
        if indexing not in ('xy', 'ij'):
            raise ValueError("Valid values for `indexing` are 'xy' and 'ij'")

        if pixelscale is None:
            raise ValueError('pixelscale must be defined to compute shift')

        x, y = 0, 0

        for tilt in self.tilt:
            x, y = tilt.shift(xs=x, ys=y, z=z, wavelength=wavelength)

        pixelscale = np.broadcast_to(pixelscale, (2,))
        out = x/pixelscale[0] * oversample, y/pixelscale[1] * oversample

        if indexing == 'ij':
            out = -out[1], out[0]

        return out

def boundary(fields):
    """
    Compute the bounding extents around a number of Field objects.

    Parameters
    ----------
    fields : :class:`~lentil.field.Field` objects
        Fields to compute boundary around

    Returns
    -------
    rmin, rmax, cmin, cmax : int
        Indices that define the extent of the shifted arrays.

    Notes
    -----
    To use the values returned by ``boundary()`` in a slice, ``rmax`` and ``cmax``
    should be increased by 1.

    See Also
    --------
    lentil.field.extent : Compute the extent of a Field

    """
    rmin, rmax, cmin, cmax = sys.maxsize, 0, sys.maxsize, 0

    for field in fields:
        frmin, frmax, fcmin, fcmax = field.extent
        rmin = frmin if frmin < rmin else rmin
        rmax = frmax if frmax > rmax else rmax
        cmin = fcmin if fcmin < cmin else cmin
        cmax = fcmax if fcmax > cmax else cmax

    return rmin, rmax, cmin, cmax

def insert(field, out, intensity=False, weight=1):
    """Insert a field into an array.

    Parameters
    ----------
    field : :class:`~lentil.field.Field`
        Field to insert into ``out``
    out : ndarray
        Array to insert ``field`` into
    intensity : bool, optional
        If True, compute intensity of ``field`` before placing it in ``out``. Default is
        False.
    weight : float, optional
        Weight to apply to ``field`` before it is inserted into ``out``
    
    Returns
    -------
    ndarray

    """
    #if indexing not in ('xy', 'ij'):
    #    raise ValueError("Valid values for `indexing` are 'xy' and 'ij'")

    if field.shape == out.shape and np.array_equal(field.offset, [0, 0]):
        field_slice = Ellipsis
        out_slice = Ellipsis
    else:

        out_shape = np.asarray(out.shape)
        field_shape = np.asarray(field.shape)
        field_offset = np.asarray(field.offset)

        #if indexing == 'xy':
        #    field_offset = -field_offset[1], field_offset[0]

        # Output coordinates of the upper left corner of the shifted data array
        field_shifted_ul = (out_shape // 2) - (field_shape // 2) + field_offset

        # Field slice indices
        field_rmin = int(0)
        field_rmax = int(field_shape[0])
        field_cmin = int(0)
        field_cmax = int(field_shape[1])

        # Output insertion slice indices
        out_rmin = int(field_shifted_ul[0])
        out_rmax = int(field_shifted_ul[0] + field_shape[0])
        out_cmin = int(field_shifted_ul[1])
        out_cmax = int(field_shifted_ul[1] + field_shape[1])

        # reconcile the field and output insertion indices
        if out_rmin < 0:
            field_rmin = -1 * out_rmin
            out_rmin = 0

        if out_rmax > out_shape[0]:
            field_rmax -= out_rmax - out_shape[0]
            out_rmax = out_shape[0]

        if out_cmin < 0:
            field_cmin = -1 * out_cmin
            out_cmin = 0

        if out_cmax > out_shape[1]:
            field_cmax -= out_cmax - out_shape[1]
            out_cmax = out_shape[1]

        out_slice = slice(out_rmin, out_rmax), slice(out_cmin, out_cmax)
        field_slice = slice(field_rmin, field_rmax), slice(field_cmin, field_cmax)

    if intensity:
        out[out_slice] += (np.abs(field.data[field_slice]**2) * weight)
    else:
        out[out_slice] += (field.data[field_slice] * weight)
    return out


def merge(a, b, enforce_overlap=True):
    """
    Merge two Fields

    Parameters
    ----------
    a, b : :class:`~lentil.field.Field` objects
        Fields to merge
    enforce_overlap : bool, optional
        If True (default), a ``ValueError`` is raised if ``a`` and ``b`` do not overlap

    Returns
    -------
    out : :class:`~lentil.field.Field`
        New ``Field`` that represents the union of ``a`` and ``b``
    """
    if enforce_overlap and not overlap((a, b)):
        raise ValueError("Can't merge non-overlapping fields")

    return _merge((a, b))
# TODO: throw an error if either field has tilts


def _merge(fields):
    """
    Merge fields into a new Field regardless of whether the supplied fields 
    overlap in space
    """
    if not np.all([f.pixelscale == fields[0].pixelscale for f in fields]):
        raise ValueError("Can't merge: pixelscales must be equal")

    out = np.zeros(_merge_shape(fields), dtype=complex)
    slices = _merge_slices(fields)
    for field, slc in zip(fields, slices):
        out[slc] += field.data

    return Field(data=out,
                 pixelscale=fields[0].pixelscale,
                 offset=_merge_offset(fields))


def _merge_shape(fields):
    """
    Return the shape required to hold merged fields
    """
    rmin, rmax, cmin, cmax = boundary(fields)
    # faster than np.any([rmin, rmax, cmin, cmax])
    if rmin == 0 and rmax == 0 and cmin == 0 and cmax == 0:
        return ()
    else:
        return rmax - rmin + 1, cmax - cmin + 1


def _merge_slices(fields):
    """
    Return the slices in a merged array for field insertion
    """
    rmin, rmax, cmin, cmax = boundary(fields)
    out = []
    # faster than np.any([rmin, rmax, cmin, cmax])
    if rmin == 0 and rmax == 0 and cmin == 0 and cmax == 0:
        out.append(Ellipsis)
    else:
        for field in fields:
            frmin, frmax, fcmin, fcmax = field.extent
            row = slice(frmin-rmin, frmax-rmin+1)
            col = slice(fcmin-cmin, fcmax-cmin+1)
            out.append((row, col))
    return out


def _merge_offset(fields):
    """
    Return the new offset of a merged field
    """
    rmin, rmax, cmin, cmax = boundary(fields)
    nrow = rmax - rmin + 1
    ncol = cmax - cmin + 1
    return rmin + nrow//2, cmin + ncol//2


def overlap(fields):
    """
    True if two Fields overlap, otherwise False

    Parameters
    ----------
    a, b : `~lentil.Field`
        Input fields.

    Returns
    -------
    overlap : bool
    
    """
    if len(fields) == 2:
        return lentil.extent.intersect(fields[0].extent, fields[1].extent)
    else:
        fields = _reduce(fields)
        if len(fields) > 1:
            return False
        else:
            return True

    
def reduce(fields):
    """
    Reduce a number of Fields into a potentially smaller disjoint set where 
    overlapping Fields are merged.

    Parameters
    ----------
    fields :class:`~lentil.field.Field` objects
        Fields to reduce

    Returns
    -------
    fields : list
        List of :class:`~lentil.field.Field` objects
    
    """
    fields = _reduce(fields)

    out = []

    for f in fields:
        if len(f['field']) > 1:
            out.append(_merge(f['field']))
        else:
            out.append(f['field'][0])

    return out


def _reduce(fields):
    """
    Function to identify overlapping fields and reorganize them into
    a dict for easier merging.
    """
    fields = [{'field': [f], 'extent': f.extent} for f in fields]
    return _disjoint(fields)


def _disjoint(fields):
    """
    Return fields as a disjoint set.
    """
    for m, n in combinations(range(len(fields)), 2):
        if lentil.extent.intersect(fields[m]['extent'], fields[n]['extent']):
            fields[m]['field'].extend(fields[n]['field'])
            fields[m]['extent'] = boundary(fields[m]['field'])
            fields.pop(n)
            return _disjoint(fields)
    return fields


def _mul_broadcast(a_data, a_offset, b_data, b_offset):
    """
    Broadcast for multiplication. 

    If one array is a scalar, it is broadcast to a compatible shape with the
    other array. As a part of this operation, the broadcasted field inherits
    the larger array's offset
    """
    # if one array is a scalar, it is broadcast to a compatible shape
    # with the other array. as a part of this operation, the broadcasted
    # Field inherits the other Field's offset as well
    #a_data, a_offset = a.data, a.offset
    #b_data, b_offset = b.data, b.offset
    if a_data.shape != b_data.shape:
        if a_data.size == 1:
            a_data = np.broadcast_to(a_data, b_data.shape)
            a_offset = b_offset
        if b_data.size == 1:
            b_data = np.broadcast_to(b_data, a_data.shape)
            b_offset = a_offset
    return a_data, a_offset, b_data, b_offset



# Functions for working with extent (rmin, rmax, cmin, cmax) data

import numpy as np

def array_extent(shape, shift, parent_shape=None):
    """Compute the extent of a shifted array.

    Parameters
    ----------
    shape : (2,) array_like
        Array shape
    shift : (2,) array_like
        Array shift in (r, c)
    parent_shape : (2,) array like or None, optional
        Enclosing parent shape. If None, the returned extent is relative
        to the origin (0,0). If provided, the returned extent is relative
        to the upper left corner of the parent shape.

    Notes
    -----
    To use the values returned by ``extent()`` in a slice, 
    ``rmax`` and ``cmax`` should be increased by 1.
    """

    if len(shape) < 2:
        shape = (1, 1)

    rmin = int(-(shape[0]//2) + shift[0])
    cmin = int(-(shape[1]//2) + shift[1])
    rmax = int(rmin + shape[0] - 1)
    cmax = int(cmin + shape[1] - 1)

    if parent_shape is not None:
        parent_center = np.asarray(parent_shape)//2
        rmin += parent_center[0]
        rmax += parent_center[0]
        cmin += parent_center[1]
        cmax += parent_center[1]

    return rmin, rmax, cmin, cmax


def array_center(extent):
    """Compute the center of an extent

    Parameters
    ----------
    a : (4,) array_like
        Array extent (rmin, rmax, cmin, cmax)

    Returns
    -------
    tuple
    """

    rmin, rmax, cmin, cmax = extent
    nrow = rmax - rmin + 1
    ncol = cmax - cmin + 1
    return rmin + nrow//2, cmin + ncol//2


def intersect(a, b):
    """Return True if two extents intersect, otherwise False

    Parameters
    ----------
    a, b : (4,) array like
        Array extents (rmin, rmax, cmin, cmax)
    
    Returns
    -------
    bool
    """

    armin, armax, acmin, acmax = a
    brmin, brmax, bcmin, bcmax = b
    return armin <= brmax and armax >= brmin and acmin <= bcmax and acmax >= bcmin


def intersection_extent(a, b):
    """Compute the extent of two overlapping extents
    
    Parameters
    ----------
    a, b : (4,) array_like
        Array extents (rmin, rmax, cmin, cmax)

    Returns
    -------
    tuple
    """

    armin, armax, acmin, acmax = a
    brmin, brmax, bcmin, bcmax = b

    rmin, rmax = max(armin, brmin), min(armax, brmax)
    cmin, cmax = max(acmin, bcmin), min(acmax, bcmax)

    return rmin, rmax, cmin, cmax


def intersection_shape(a, b):
    """Compute the shape of two overlapping extents. If there is no
    overlap, an empty tuple is returned.

    Parameters
    ----------
    a, b : (4,) array like
        Array extents (rmin, rmax, cmin, cmax)
    
    Returns
    -------
    tuple
    """

    rmin, rmax, cmin, cmax = intersection_extent(a, b)
    nr, nc = rmax - rmin + 1, cmax - cmin + 1

    if nr <= 0 or nc <= 0:
        shape = ()
    else:
        shape = (nr, nc)
    
    return shape


def intersection_slices(a, b):
    """Compute slices of overlapping areas between two overlapping extents

    Parameters
    ----------
    a, b : (4,) array like
        Array extents (rmin, rmax, cmin, cmax)
    
    Returns
    -------
    tuples of slices
    """

    rmin, rmax, cmin, cmax = intersection_extent(a, b)

    armin, armax, acmin, acmax = a
    brmin, brmax, bcmin, bcmax = b

    arow = slice(rmin-armin, rmax-armin+1)
    acol = slice(cmin-acmin, cmax-acmin+1)
    brow = slice(rmin-brmin, rmax-brmin+1)
    bcol = slice(cmin-bcmin, cmax-bcmin+1)

    return (arow, acol), (brow, bcol)


def intersection_shift(a, b):
    """Compute the shift between two overlapping extents

    Parameters
    ----------
    a, b : (4,) array like
        Array extents (rmin, rmax, cmin, cmax)
    
    Returns
    -------
    tuple
    """
    
    rmin, rmax, cmin, cmax = intersection_extent(a, b)
    nrow = rmax - rmin + 1
    ncol = cmax - cmin + 1
    return rmin + nrow//2, cmin + ncol//2


# These tests deal specifically with ensuring sliced and offset propagations
# are equivalent to their full-sized counterparts

import numpy as np
import lentil
import lentil.helper


def test_propagate_slice_one():
    # Single monolithic aperture

    diameter = 1
    focal_length = 20
    wavelength = 500e-9
    du = 5e-6
    dx = diameter/128

    shift = np.random.uniform(low=-50, high=50, size=2).astype(int)
    amp = lentil.circle((256,256), 256//4, shift=shift)

    rho, theta = lentil.zernike_coordinates(amp, shift=shift)
    coeffs = np.random.uniform(low=-1, high=1, size=11)*100e-9
    coeffs[0:3] = 0 # no ptt
    opd = lentil.zernike_compose(amp, coeffs, rho=rho, theta=theta)

    alpha = (dx*du)/(wavelength*focal_length)
    oversample = 5
    shape = 64

    phasor = amp*np.exp(-2*1j*np.pi*opd/wavelength)
    F = lentil.fourier.dft2(phasor, alpha=alpha/oversample, shape=shape*oversample)


    slc = lentil.helper.boundary_slice(amp)
    ofst = lentil.helper.slice_offset(slc, shape=amp.shape)

    phasor = amp[slc]*np.exp(-2*1j*np.pi*opd[slc]/wavelength)
    F_slc = lentil.fourier.dft2(phasor, alpha=alpha/oversample, shape=shape*oversample, offset=ofst)

    assert np.allclose(F, F_slc)


def test_propagate_slice_multi():
    # Segmented aperture
    diameter = 1
    focal_length = 20
    wavelength = 500e-9
    du = 5e-6
    dx = diameter/102

    n = 256

    amp1 = lentil.circle((n,n), n//5, shift=(0, -0.3*n))
    slc1 = lentil.helper.boundary_slice(amp1)
    ofst1 = lentil.helper.slice_offset(slc1, shape=amp1.shape)

    rho, theta = lentil.zernike_coordinates(amp1, shift=(0, -0.3*n))
    coeffs = np.random.uniform(low=-1, high=1, size=11)*100e-9
    coeffs[0:3] = 0 # no ptt
    opd1 = lentil.zernike_compose(amp1, coeffs, rho=rho, theta=theta)


    amp2 = lentil.circle((n,n), n//5, shift=(0, .3*n))
    slc2 = lentil.helper.boundary_slice(amp2)
    ofst2 = lentil.helper.slice_offset(slc2, shape=amp2.shape)

    rho, theta = lentil.zernike_coordinates(amp2, shift=(0, 0.3*n))
    coeffs = np.random.uniform(low=-1, high=1, size=11)*100e-9
    coeffs[0:3] = 0 # no ptt
    opd2 = lentil.zernike_compose(amp2, coeffs, rho=rho, theta=theta)

    alpha = (dx*du)/(wavelength*focal_length)
    oversample = 5
    shape = 64

    amp = amp1 + amp2
    opd = opd1 + opd2

    phasor = amp*np.exp(-2*1j*np.pi*opd/wavelength)
    F = lentil.fourier.dft2(phasor, alpha=alpha/oversample, shape=shape*oversample)

    phasor1 = amp1[slc1]*np.exp(-2*1j*np.pi*opd1[slc1]/wavelength)
    F1 = lentil.fourier.dft2(phasor1, alpha=alpha/oversample, shape=shape*oversample, offset=ofst1)

    phasor2 = amp2[slc2]*np.exp(-2*1j*np.pi*opd2[slc2]/wavelength)
    F2 = lentil.fourier.dft2(phasor2, alpha=alpha/oversample, shape=shape*oversample, offset=ofst2)

    F_slc = F1 + F2

    np.allclose(F, F_slc)



import numpy as np
import lentil

def test_propagate_mask(pupil):
    p = pupil(focal_length=10, diameter=1, shape=(256,256), radius=120, coeffs=None)

    mask = lentil.rectangle((256,256), 128,128, shift=(20, -30), antialias=False)

    w = lentil.Wavefront(650e-9)
    w = p.multiply(w)
    w = lentil.propagate_dft(w, shape=128, pixelscale=5e-6, oversample=2)
    psf = w.intensity

    w_mask = lentil.Wavefront(650e-9)
    w_mask = p.multiply(w_mask)
    w_mask = lentil.propagate_dft(w_mask, shape=128, pixelscale=5e-6, oversample=2, mask=mask)
    psf_mask = w_mask.intensity

    assert np.allclose(psf_mask, psf*mask)


def test_propagate_mask_tilt_analytic(pupil):
    p = pupil(focal_length=10, diameter=1, shape=(256,256), radius=120, coeffs=[0,1e-6, 2e-6])

    mask = lentil.rectangle((256,256), 64, 64, shift=(20, -30), antialias=False)

    w = lentil.Wavefront(650e-9)
    w = p.multiply(w)
    w = lentil.propagate_dft(w, shape=128, pixelscale=5e-6, oversample=2)
    psf = w.intensity

    w_mask = lentil.Wavefront(650e-9)
    w_mask = p.multiply(w_mask)
    w_mask = lentil.propagate_dft(w_mask, shape=128, pixelscale=5e-6, oversample=2, mask=mask)
    psf_mask = w_mask.intensity

    assert np.allclose(psf_mask, psf*mask)

